from __future__ import annotations

import functools
import logging
from collections.abc import Iterator, Sequence
from pathlib import Path
from typing import TYPE_CHECKING, ClassVar, Generic, Literal, cast, overload

import torch
from lightning.fabric.utilities.apply_func import move_data_to_device
from lightning.pytorch.callbacks import BasePredictionWriter
from typing_extensions import TypeVar, final, override

from .base import CallbackConfigBase, CallbackMetadataConfig, callback_registry

if TYPE_CHECKING:
    from ..model.base import IndividualSample


log = logging.getLogger(__name__)


@final
@callback_registry.register
class DistributedPredictionWriterConfig(CallbackConfigBase):
    metadata: ClassVar[CallbackMetadataConfig] = CallbackMetadataConfig(
        enabled_for_barebones=True
    )
    """Metadata for the callback."""

    name: Literal["distributed_prediction_writer"] = "distributed_prediction_writer"

    dirpath: Path | None = None
    """Directory to save the predictions to. If None, will use the default directory."""

    move_to_cpu_on_save: bool = True
    """Whether to move the predictions to CPU before saving. Default is True."""

    save_raw: bool = True
    """Whether to save the raw predictions."""

    save_processed: bool = True
    """Whether to process and save the predictions.

    "Processing" means that the model's batched predictions are split into individual predictions
    and saved as a list of tensors.
    """

    @override
    def create_callbacks(self, trainer_config):
        if (dirpath := self.dirpath) is None:
            dirpath = trainer_config.directory.resolve_subdirectory(
                trainer_config.id, "predictions"
            )

        yield DistributedPredictionWriter(self, dirpath)


def _move_and_save(data, path: Path, move_to_cpu: bool):
    if move_to_cpu:
        data = move_data_to_device(data, "cpu")

    # Save the data to the specified path
    torch.save(data, path)


class DistributedPredictionWriter(BasePredictionWriter):
    def __init__(
        self,
        config: DistributedPredictionWriterConfig,
        output_dir: Path,
    ):
        self.config = config

        super().__init__(write_interval="batch")

        self.output_dir = output_dir

    @override
    def write_on_batch_end(
        self,
        trainer,
        pl_module,
        prediction,
        batch_indices,
        batch,
        batch_idx,
        dataloader_idx,
    ):
        save = functools.partial(
            _move_and_save,
            move_to_cpu=self.config.move_to_cpu_on_save,
        )

        # Regular, unstructured writing.
        if self.config.save_raw:
            output_dir = (
                self.output_dir
                / "raw"
                / f"dataloader_{dataloader_idx}"
                / f"rank_{trainer.global_rank}"
                / f"batch_{batch_idx}"
            )
            output_dir.mkdir(parents=True, exist_ok=True)
            save(prediction, output_dir / "predictions.pt")
            save(batch, output_dir / "batch.pt")
            save(batch_indices, output_dir / "batch_indices.pt")

        if self.config.save_processed:
            # Processed writing.
            from ..model.base import LightningModuleBase

            if not isinstance(pl_module, LightningModuleBase):
                raise ValueError(
                    "The model must be a subclass of LightningModuleBase to use the distributed prediction writer."
                )

            output_dir = self.output_dir / "processed" / f"dataloader_{dataloader_idx}"
            output_dir.mkdir(parents=True, exist_ok=True)

            # Split into individual predictions
            assert batch_indices is not None, (
                "Batch indices must be provided for processed writing."
            )
            for sample in pl_module.split_batched_predictions(
                batch, prediction, batch_indices
            ):
                sample = {
                    **sample,
                    "global_rank": trainer.global_rank,
                    "world_size": trainer.world_size,
                    "is_global_zero": trainer.is_global_zero,
                }
                save(sample, output_dir / f"{sample['index']}.pt")


SampleT = TypeVar(
    "SampleT",
    bound="IndividualSample",
    default="IndividualSample",
    infer_variance=True,
)


class DistributedPredictionReader(Sequence[SampleT], Generic[SampleT]):
    def __init__(self, output_dir: Path):
        self.output_dir = output_dir

    @override
    def __len__(self) -> int:
        return len(list(self.output_dir.glob("*.pt")))

    @overload
    def __getitem__(self, index: int) -> SampleT: ...

    @overload
    def __getitem__(self, index: slice) -> list[SampleT]: ...

    @override
    def __getitem__(self, index: int | slice) -> SampleT | list[SampleT]:
        if isinstance(index, slice):
            # Handle slice indexing
            indices = range(*index.indices(len(self)))
            return [self.__getitem__(i) for i in indices]

        # Handle integer indexing
        path = self.output_dir / f"{index}.pt"
        if not path.exists():
            raise FileNotFoundError(f"File {path} does not exist.")

        sample = cast(SampleT, torch.load(path))
        return sample

    @override
    def __iter__(self) -> Iterator[SampleT]:
        for i in range(len(self)):
            yield self[i]



from __future__ import annotations

import logging
from collections.abc import Mapping
from typing import Literal, cast

import torch
from lightning.pytorch import LightningModule
from lightning.pytorch.callbacks import Callback
from lightning.pytorch.utilities.types import (
    LRSchedulerConfigType,
    LRSchedulerTypeUnion,
)
from typing_extensions import Protocol, final, override, runtime_checkable

from .base import CallbackConfigBase, callback_registry

log = logging.getLogger(__name__)


@final
@callback_registry.register
class RLPSanityChecksCallbackConfig(CallbackConfigBase):
    """
    If enabled, will do some sanity checks if the `ReduceLROnPlateau` scheduler is used:
        - If the ``interval`` is step, it makes sure that validation is called every ``frequency`` steps.
        - If the ``interval`` is epoch, it makes sure that validation is called every ``frequency`` epochs.
    """

    name: Literal["rlp_sanity_checks"] = "rlp_sanity_checks"

    enabled: bool = True
    """Whether to enable ReduceLRPlateau sanity checks."""

    on_error: Literal["warn", "error"] = "error"
    """What to do when a sanity check fails."""

    def __bool__(self):
        return self.enabled

    def create_callbacks(self, trainer_config):
        if not self:
            return

        yield RLPSanityChecksCallback(self)


class RLPSanityChecksCallback(Callback):
    @override
    def __init__(self, config: RLPSanityChecksCallbackConfig):
        super().__init__()

        self.config = config
        del config

    @override
    def on_train_start(self, trainer, pl_module):
        # If we're in PL's "sanity check" mode, we don't need to run this check
        if trainer.sanity_checking:
            return

        # If the sanity check is disabled, return.
        if not self.config:
            return

        # If no lr schedulers, return.
        if not trainer.lr_scheduler_configs:
            return

        errors: list[str] = []
        disable_message = (
            "Otherwise, set `config.trainer.sanity_checking.reduce_lr_on_plateau = None` "
            "to disable this sanity check."
        )

        for lr_scheduler_config in trainer.lr_scheduler_configs:
            if not lr_scheduler_config.reduce_on_plateau:
                continue

            match lr_scheduler_config.interval:
                case "epoch":
                    # we need to make sure that the trainer runs val every `frequency` epochs

                    # If `trainer.check_val_every_n_epoch` is None, then Lightning
                    # will run val every `int(trainer.val_check_interval)` steps.
                    # So, first we need to make sure that `trainer.val_check_interval` is not None first.
                    if trainer.check_val_every_n_epoch is None:
                        errors.append(
                            "Trainer is not running validation at epoch intervals "
                            "(i.e., `trainer.check_val_every_n_epoch` is None) but "
                            f"a ReduceLRPlateau scheduler with interval={lr_scheduler_config.interval} is used."
                            f"Please set `config.trainer.check_val_every_n_epoch={lr_scheduler_config.frequency}`. "
                            + disable_message
                        )

                    # Second, we make sure that the trainer runs val at least every `frequency` epochs
                    if (
                        trainer.check_val_every_n_epoch is not None
                        and lr_scheduler_config.frequency
                        % trainer.check_val_every_n_epoch
                        != 0
                    ):
                        errors.append(
                            f"Trainer is not running validation every {lr_scheduler_config.frequency} epochs but "
                            f"a ReduceLRPlateau scheduler with interval={lr_scheduler_config.interval} and frequency={lr_scheduler_config.frequency} is used."
                            f"Please set `config.trainer.check_val_every_n_epoch` to a multiple of {lr_scheduler_config.frequency}. "
                            + disable_message
                        )

                case "step":
                    # In this case, we need to make sure that the trainer runs val at step intervals
                    # that are multiples of `frequency`.

                    # First, we make sure that validation is run at step intervals
                    if trainer.check_val_every_n_epoch is not None:
                        errors.append(
                            "Trainer is running validation at epoch intervals "
                            "(i.e., `trainer.check_val_every_n_epoch` is not None) but "
                            f"a ReduceLRPlateau scheduler with interval={lr_scheduler_config.interval} is used."
                            "Please set `config.trainer.check_val_every_n_epoch=None` "
                            f"and `config.trainer.val_check_interval={lr_scheduler_config.frequency}`. "
                            + disable_message
                        )

                    # Second, we make sure `trainer.val_check_interval` is an integer
                    if not isinstance(trainer.val_check_interval, int):
                        errors.append(
                            f"Trainer is not running validation at step intervals "
                            f"(i.e., `trainer.val_check_interval` is not an integer) but "
                            f"a ReduceLRPlateau scheduler with interval={lr_scheduler_config.interval} is used."
                            "Please set `config.trainer.val_check_interval=None` "
                            f"and `config.trainer.val_check_interval={lr_scheduler_config.frequency}`. "
                            + disable_message
                        )

                    # Third, we make sure that the trainer runs val at least every `frequency` steps
                    if (
                        isinstance(trainer.val_check_interval, int)
                        and trainer.val_check_interval % lr_scheduler_config.frequency
                        != 0
                    ):
                        errors.append(
                            f"Trainer is not running validation every {lr_scheduler_config.frequency} steps but "
                            f"a ReduceLRPlateau scheduler with interval={lr_scheduler_config.interval} and frequency={lr_scheduler_config.frequency} is used."
                            "Please set `config.trainer.val_check_interval` "
                            f"to a multiple of {lr_scheduler_config.frequency}. "
                            + disable_message
                        )

                case _:
                    pass

        if not errors:
            return

        message = (
            "ReduceLRPlateau sanity checks failed with the following errors:\n"
            + "\n".join(errors)
        )
        match self.config.on_error:
            case "warn":
                log.warning(message)
            case "error":
                raise ValueError(message)
            case _:
                pass


@runtime_checkable
class CustomRLPImplementation(Protocol):
    __reduce_lr_on_plateau__: bool


class RLPSanityCheckModuleMixin(LightningModule):
    def reduce_lr_on_plateau_config(
        self,
        lr_scheduler: LRSchedulerTypeUnion | LRSchedulerConfigType,
    ) -> LRSchedulerConfigType:
        if (trainer := self._trainer) is None:
            raise RuntimeError(
                "Could not determine the frequency of ReduceLRPlateau scheduler "
                "because `self.trainer` is None."
            )

        # First, resolve the LR scheduler from the provided config.
        lr_scheduler_config: LRSchedulerConfigType
        match lr_scheduler:
            case Mapping():
                lr_scheduler_config = cast(LRSchedulerConfigType, lr_scheduler)
            case _:
                lr_scheduler_config = {"scheduler": lr_scheduler}

        # Make sure the scheduler is a ReduceLRPlateau scheduler. Otherwise, warn the user.
        if (
            not isinstance(
                lr_scheduler_config["scheduler"],
                torch.optim.lr_scheduler.ReduceLROnPlateau,
            )
        ) and (
            not isinstance(lr_scheduler_config["scheduler"], CustomRLPImplementation)
            or not lr_scheduler_config["scheduler"].__reduce_lr_on_plateau__
        ):
            log.warning(
                "`reduce_lr_on_plateau_config` should only be used with a ReduceLRPlateau scheduler. "
                f"The provided scheduler, {lr_scheduler_config['scheduler']}, does not subclass "
                "`torch.optim.lr_scheduler.ReduceLROnPlateau`. "
                "Please ensure that the scheduler is a ReduceLRPlateau scheduler. "
                "If you are using a custom ReduceLRPlateau scheduler implementation, "
                "please either (1) make sure that it subclasses `torch.optim.lr_scheduler.ReduceLROnPlateau`, "
                "or (2) set the scheduler's `__reduce_lr_on_plateau__` attribute to `True`."
            )

        # If trainer.check_val_every_n_epoch is an integer, then we run val at epoch intervals.
        if trainer.check_val_every_n_epoch is not None:
            return {
                "reduce_on_plateau": True,
                "interval": "epoch",
                "frequency": trainer.check_val_every_n_epoch,
                **lr_scheduler_config,
            }

        # Otherwise, we run val at step intervals.
        if not isinstance(trainer.val_check_batch, int):
            raise ValueError(
                "Could not determine the frequency of ReduceLRPlateau scheduler "
                f"because {trainer.val_check_batch=} is not an integer."
            )

        return {
            "reduce_on_plateau": True,
            "interval": "step",
            "frequency": trainer.val_check_batch,
            **lr_scheduler_config,
        }



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from collections.abc import Callable, Iterable, Mapping, Sequence
from pathlib import Path
from typing import Any, Generic, Literal, TypedDict, cast

import nshconfig as C
import torch
import torch.distributed
from lightning.pytorch import LightningModule
from lightning.pytorch.profilers import PassThroughProfiler, Profiler
from lightning.pytorch.utilities.model_helpers import is_overridden
from lightning.pytorch.utilities.rank_zero import rank_zero_warn
from typing_extensions import Never, TypeVar, deprecated, override

from ..callbacks.rlp_sanity_checks import RLPSanityCheckModuleMixin
from .mixins.callback import CallbackModuleMixin
from .mixins.debug import DebugModuleMixin
from .mixins.logger import LoggerLightningModuleMixin

log = logging.getLogger(__name__)

THparams = TypeVar("THparams", bound=C.Config, infer_variance=True)


T = TypeVar("T", infer_variance=True)

ReduceOpStr = Literal[
    "avg",
    "mean",
    "band",
    "bor",
    "bxor",
    "max",
    "min",
    "premul_sum",
    "product",
    "sum",
]
VALID_REDUCE_OPS = (
    "avg",
    "mean",
    "band",
    "bor",
    "bxor",
    "max",
    "min",
    "premul_sum",
    "product",
    "sum",
)


class IndividualSample(TypedDict):
    """
    A dictionary that contains the individual sample.
    This is used to split the batched predictions into individual predictions.
    """

    index: int
    """The index of the sample in the batch."""

    batch: Any
    """The batch to split."""

    prediction: Any
    """The batched prediction to split."""


def default_split_batched_predictions(
    batch: Any,
    prediction: Any,
    batch_indices: Sequence[Any],
) -> Iterable[IndividualSample]:
    """
    Splits the batched predictions into a list of individual predictions.
    Args:
        batch: The batch to split.
        prediction: The batched prediction to split.
        batch_indices: The indices of the batches.
    Returns:
        A tuple of two sequences: the corresponding batches and the individual predictions.
    """
    import torch.utils._pytree as tree

    for i, global_idx in enumerate(batch_indices):

        def _verify_and_index(x: torch.Tensor):
            # Make sure dim 0 length is equal to the batch size,
            # otherwise we can't index it and should prompt
            # the user to implement a splitter
            if x.shape[0] != len(batch_indices):
                raise ValueError(
                    f"Batch size {x.shape[0]} does not match the number of batch indices {len(batch_indices)}. "
                    "Please implement a custom `split_batched_predictions` method in your LightningModuleBase class."
                )

            return x[i]

        # Create a dictionary for each sample
        yield IndividualSample(
            index=global_idx,
            batch=tree.tree_map(_verify_and_index, batch),
            prediction=tree.tree_map(_verify_and_index, prediction),
        )


class LightningModuleBase(
    DebugModuleMixin,
    RLPSanityCheckModuleMixin,
    LoggerLightningModuleMixin,
    CallbackModuleMixin,
    LightningModule,
    ABC,
    Generic[THparams],
):
    # region Profiler
    @property
    def profiler(self) -> Profiler:
        if (trainer := self._trainer) is None:
            raise RuntimeError("trainer is not defined")

        if not hasattr(trainer, "profiler"):
            raise RuntimeError("trainer does not have profiler")

        if (profiler := getattr(trainer, "profiler")) is None:
            profiler = PassThroughProfiler()

        return profiler

    # endregion

    # region Distributed
    def all_gather_object(
        self,
        object: T,
        group: torch.distributed.ProcessGroup | None = None,
    ) -> list[T]:
        if (
            not torch.distributed.is_available()
            or not torch.distributed.is_initialized()
        ):
            return [object]

        object_list = [cast(T, None) for _ in range(self.trainer.world_size)]
        torch.distributed.all_gather_object(object_list, object, group=group)
        return object_list

    def barrier(self, name: str | None = None):
        return self.trainer.strategy.barrier(name=name)

    def reduce(
        self,
        tensor: torch.Tensor,
        reduce_op: torch.distributed.ReduceOp.RedOpType | ReduceOpStr,
        group: Any | None = None,
    ) -> torch.Tensor:
        if isinstance(reduce_op, str):
            # validate reduce_op
            if reduce_op not in VALID_REDUCE_OPS:
                raise ValueError(
                    f"reduce_op must be one of {VALID_REDUCE_OPS}, got {reduce_op}"
                )

        return self.trainer.strategy.reduce(tensor, group=group, reduce_op=reduce_op)

    # endregion

    # Our own custom __repr__ method.
    # Torch's __repr__ method is too verbose and doesn't provide any useful information.
    @override
    def __repr__(self):
        parts: list[str] = []
        parts.append(f"hparams={repr(self.hparams)}")
        parts.append(f"device={self.device}")
        if self.debug:
            parts.append("debug=True")

        parts_str = ", ".join(parts)
        return f"{self.__class__.__name__}({parts_str})"

    @property
    @override
    def hparams(self) -> THparams:  # pyright: ignore[reportIncompatibleMethodOverride]
        return cast(THparams, super().hparams)

    @property
    @override
    def hparams_initial(self) -> THparams:  # pyright: ignore[reportIncompatibleMethodOverride]
        hparams = cast(THparams, super().hparams_initial)
        return hparams

    @property
    @deprecated("Use `hparams` instead")
    def config(self):
        return cast(Never, self.hparams)

    @classmethod
    @abstractmethod
    def hparams_cls(cls) -> type[THparams]: ...

    @override
    def __init__(self, hparams: THparams | Mapping[str, Any]):
        super().__init__()

        # Validate and save hyperparameters
        hparams_cls = self.hparams_cls()
        if isinstance(hparams, Mapping):
            hparams = hparams_cls.model_validate(hparams)
        elif not isinstance(hparams, hparams_cls):
            raise TypeError(
                f"Expected hparams to be either a Mapping or an instance of {hparams_cls}, got {type(hparams)}"
            )
        hparams = hparams.model_deep_validate()
        self.save_hyperparameters(hparams)

    def zero_loss(self):
        """
        Returns a loss tensor with the value 0.
        It multiples each weight by 0 and returns the sum, so we don't run into issues with ununsed parameters in DDP.
        """
        loss = sum((0.0 * v).sum() for v in self.parameters() if v.requires_grad)
        loss = cast(torch.Tensor, loss)
        return loss

    def split_batched_predictions(
        self,
        batch: Any,
        prediction: Any,
        batch_indices: Sequence[Any],
    ) -> Iterable[IndividualSample]:
        """
        Splits the batched predictions into a list of individual predictions.
        Args:
            batch: The batch to split.
            prediction: The batched prediction to split.
            batch_indices: The indices of the batches.
        Returns:
            A tuple of two sequences: the corresponding batches and the individual predictions.
        """
        return default_split_batched_predictions(batch, prediction, batch_indices)

    @override
    @classmethod
    def load_from_checkpoint(cls, *args, **kwargs) -> Never:
        raise ValueError("This method is not supported. Use `from_checkpoint` instead.")

    @classmethod
    def hparams_from_checkpoint(
        cls,
        ckpt_or_path: dict[str, Any] | str | Path,
        /,
        strict: bool | None = None,
        *,
        update_hparams: Callable[[THparams], THparams] | None = None,
        update_hparams_dict: Callable[[dict[str, Any]], dict[str, Any]] | None = None,
    ):
        if isinstance(ckpt_or_path, dict):
            ckpt = ckpt_or_path
        else:
            ckpt = torch.load(ckpt_or_path, map_location="cpu")

        if (hparams := ckpt.get(cls.CHECKPOINT_HYPER_PARAMS_KEY)) is None:
            raise ValueError(
                f"The checkpoint does not contain hyperparameters. It must contain the key '{cls.CHECKPOINT_HYPER_PARAMS_KEY}'."
            )
        if update_hparams_dict is not None:
            hparams = update_hparams_dict(hparams)

        hparams = cls.hparams_cls().model_validate(hparams, strict=strict)
        if update_hparams is not None:
            hparams = update_hparams(hparams)

        return hparams

    @classmethod
    def from_checkpoint(
        cls,
        ckpt_or_path: dict[str, Any] | str | Path,
        /,
        strict: bool | None = None,
        map_location: torch.serialization.MAP_LOCATION = None,
        *,
        update_hparams: Callable[[THparams], THparams] | None = None,
        update_hparams_dict: Callable[[dict[str, Any]], dict[str, Any]] | None = None,
    ):
        # Load checkpoint
        if isinstance(ckpt_or_path, Mapping):
            ckpt = ckpt_or_path
        else:
            ckpt = torch.load(ckpt_or_path, map_location=map_location)

        # Load hyperparameters from checkpoint
        hparams = cls.hparams_from_checkpoint(
            ckpt,
            strict=strict,
            update_hparams=update_hparams,
            update_hparams_dict=update_hparams_dict,
        )

        # Load model from checkpoint
        model = cls(hparams)

        # Load model state from checkpoint
        if (
            model._strict_loading is not None
            and strict is not None
            and strict != model.strict_loading
        ):
            raise ValueError(
                f"You set `.load_from_checkpoint(..., strict={strict!r})` which is in conflict with"
                f" `{cls.__name__}.strict_loading={model.strict_loading!r}. Please set the same value for both of them."
            )
        strict = model.strict_loading if strict is None else strict

        if is_overridden("configure_model", model):
            model.configure_model()

        # give model a chance to load something
        model.on_load_checkpoint(ckpt)

        # load the state_dict on the model automatically

        keys = model.load_state_dict(ckpt["state_dict"], strict=strict)

        if not strict:
            if keys.missing_keys:
                rank_zero_warn(
                    f"Found keys that are in the model state dict but not in the checkpoint: {keys.missing_keys}"
                )
            if keys.unexpected_keys:
                rank_zero_warn(
                    f"Found keys that are not in the model state dict but in the checkpoint: {keys.unexpected_keys}"
                )

        return model



from __future__ import annotations

from .mlp import MLP as MLP
from .mlp import MLPConfig as MLPConfig
from .mlp import ResidualSequential as ResidualSequential
from .module_dict import TypedModuleDict as TypedModuleDict
from .module_list import TypedModuleList as TypedModuleList
from .nonlinearity import ELUNonlinearityConfig as ELUNonlinearityConfig
from .nonlinearity import GELUNonlinearityConfig as GELUNonlinearityConfig
from .nonlinearity import LeakyReLUNonlinearityConfig as LeakyReLUNonlinearityConfig
from .nonlinearity import MishNonlinearityConfig as MishNonlinearityConfig
from .nonlinearity import NonlinearityConfig as NonlinearityConfig
from .nonlinearity import NonlinearityConfigBase as NonlinearityConfigBase
from .nonlinearity import PReLUConfig as PReLUConfig
from .nonlinearity import ReLUNonlinearityConfig as ReLUNonlinearityConfig
from .nonlinearity import SigmoidNonlinearityConfig as SigmoidNonlinearityConfig
from .nonlinearity import SiLUNonlinearityConfig as SiLUNonlinearityConfig
from .nonlinearity import SoftmaxNonlinearityConfig as SoftmaxNonlinearityConfig
from .nonlinearity import SoftplusNonlinearityConfig as SoftplusNonlinearityConfig
from .nonlinearity import SoftsignNonlinearityConfig as SoftsignNonlinearityConfig
from .nonlinearity import SwishNonlinearityConfig as SwishNonlinearityConfig
from .nonlinearity import TanhNonlinearityConfig as TanhNonlinearityConfig
from .rng import RNGConfig as RNGConfig
from .rng import rng_context as rng_context



from __future__ import annotations

from typing import Annotated

from typing_extensions import TypeAliasType

from . import checkpoint as checkpoint
from .actsave import ActSaveCallback as ActSaveCallback
from .actsave import ActSaveConfig as ActSaveConfig
from .base import CallbackConfigBase as CallbackConfigBase
from .base import callback_registry as callback_registry
from .checkpoint import BestCheckpointCallback as BestCheckpointCallback
from .checkpoint import BestCheckpointCallbackConfig as BestCheckpointCallbackConfig
from .checkpoint import LastCheckpointCallback as LastCheckpointCallback
from .checkpoint import LastCheckpointCallbackConfig as LastCheckpointCallbackConfig
from .checkpoint import OnExceptionCheckpointCallback as OnExceptionCheckpointCallback
from .checkpoint import (
    OnExceptionCheckpointCallbackConfig as OnExceptionCheckpointCallbackConfig,
)
from .debug_flag import DebugFlagCallback as DebugFlagCallback
from .debug_flag import DebugFlagCallbackConfig as DebugFlagCallbackConfig
from .directory_setup import DirectorySetupCallback as DirectorySetupCallback
from .directory_setup import (
    DirectorySetupCallbackConfig as DirectorySetupCallbackConfig,
)
from .distributed_prediction_writer import (
    DistributedPredictionWriter as DistributedPredictionWriter,
)
from .distributed_prediction_writer import (
    DistributedPredictionWriterConfig as DistributedPredictionWriterConfig,
)
from .early_stopping import EarlyStoppingCallback as EarlyStoppingCallback
from .early_stopping import EarlyStoppingCallbackConfig as EarlyStoppingCallbackConfig
from .ema import EMACallback as EMACallback
from .ema import EMACallbackConfig as EMACallbackConfig
from .finite_checks import FiniteChecksCallback as FiniteChecksCallback
from .finite_checks import FiniteChecksCallbackConfig as FiniteChecksCallbackConfig
from .gradient_skipping import GradientSkippingCallback as GradientSkippingCallback
from .gradient_skipping import (
    GradientSkippingCallbackConfig as GradientSkippingCallbackConfig,
)
from .interval import EpochIntervalCallback as EpochIntervalCallback
from .interval import IntervalCallback as IntervalCallback
from .interval import StepIntervalCallback as StepIntervalCallback
from .log_epoch import LogEpochCallback as LogEpochCallback
from .log_epoch import LogEpochCallbackConfig as LogEpochCallbackConfig
from .lr_monitor import LearningRateMonitor as LearningRateMonitor
from .lr_monitor import LearningRateMonitorConfig as LearningRateMonitorConfig
from .metric_validation import MetricValidationCallback as MetricValidationCallback
from .metric_validation import (
    MetricValidationCallbackConfig as MetricValidationCallbackConfig,
)
from .norm_logging import NormLoggingCallback as NormLoggingCallback
from .norm_logging import NormLoggingCallbackConfig as NormLoggingCallbackConfig
from .print_table import PrintTableMetricsCallback as PrintTableMetricsCallback
from .print_table import (
    PrintTableMetricsCallbackConfig as PrintTableMetricsCallbackConfig,
)
from .rlp_sanity_checks import RLPSanityChecksCallback as RLPSanityChecksCallback
from .rlp_sanity_checks import (
    RLPSanityChecksCallbackConfig as RLPSanityChecksCallbackConfig,
)
from .shared_parameters import SharedParametersCallback as SharedParametersCallback
from .shared_parameters import (
    SharedParametersCallbackConfig as SharedParametersCallbackConfig,
)
from .timer import EpochTimerCallback as EpochTimerCallback
from .timer import EpochTimerCallbackConfig as EpochTimerCallbackConfig
from .wandb_upload_code import WandbUploadCodeCallback as WandbUploadCodeCallback
from .wandb_upload_code import (
    WandbUploadCodeCallbackConfig as WandbUploadCodeCallbackConfig,
)
from .wandb_watch import WandbWatchCallback as WandbWatchCallback
from .wandb_watch import WandbWatchCallbackConfig as WandbWatchCallbackConfig

CallbackConfig = TypeAliasType(
    "CallbackConfig",
    Annotated[CallbackConfigBase, callback_registry.DynamicResolution()],
)



from __future__ import annotations

from typing import Any, Literal

from typing_extensions import override

from ...util.config.dtype import DTypeConfig
from .base import PluginConfigBase, plugin_registry


@plugin_registry.register
class MixedPrecisionPluginConfig(PluginConfigBase):
    name: Literal["mixed_precision"] = "mixed_precision"

    precision: Literal["16-mixed", "bf16-mixed"]
    """Whether to use ``torch.float16`` (``'16-mixed'``) or ``torch.bfloat16`` (``'bf16-mixed'``)."""

    device: str
    """The device for ``torch.autocast``."""

    @override
    def create_plugin(self, trainer_config):
        from lightning.pytorch.plugins.precision.amp import MixedPrecision

        return MixedPrecision(self.precision, self.device)


@plugin_registry.register
class BitsandbytesPluginConfig(PluginConfigBase):
    name: Literal["bitsandbytes_precision"] = "bitsandbytes_precision"

    mode: Literal["nf4", "nf4-dq", "fp4", "fp4-dq", "int8", "int8-training"]
    """The quantization mode to use."""

    dtype: DTypeConfig | None = None
    """The compute dtype to use."""

    ignore_modules: set[str] | None = None
    """The submodules whose Linear layers should not be replaced.

    This might be desirable for numerical stability. The string will be checked
    as a prefix, so a value like "transformer.blocks" will ignore all linear
    layers in all of the transformer blocks.
    """

    @override
    def create_plugin(self, trainer_config):
        from lightning.pytorch.plugins.precision.bitsandbytes import (
            BitsandbytesPrecision,
        )

        return BitsandbytesPrecision(
            mode=self.mode,
            dtype=self.dtype.torch_dtype if self.dtype is not None else None,
            ignore_modules=self.ignore_modules,
        )


@plugin_registry.register
class DeepSpeedPluginConfig(PluginConfigBase):
    name: Literal["deepspeed_precision"] = "deepspeed_precision"

    precision: Literal["16-true", "bf16-true", "16-mixed", "bf16-mixed", "32-true"]
    """Full precision (32-true), half precision (16-true, bf16-true) or
    mixed precision (16-mixed, bf16-mixed)."""

    @override
    def create_plugin(self, trainer_config):
        from lightning.pytorch.plugins.precision.deepspeed import DeepSpeedPrecision

        return DeepSpeedPrecision(precision=self.precision)


@plugin_registry.register
class DoublePrecisionPluginConfig(PluginConfigBase):
    name: Literal["double_precision"] = "double_precision"

    precision: Literal["64-true"] = "64-true"
    """Plugin for training with double (``torch.float64``) precision."""

    @override
    def create_plugin(self, trainer_config):
        from lightning.pytorch.plugins.precision.double import DoublePrecision

        return DoublePrecision()


@plugin_registry.register
class FSDPPrecisionPluginConfig(PluginConfigBase):
    name: Literal["fsdp_precision"] = "fsdp_precision"

    precision: Literal["16-true", "bf16-true", "16-mixed", "bf16-mixed", "32-true"]
    """Full precision (32-true), half precision (16-true, bf16-true) or
    mixed precision (16-mixed, bf16-mixed)."""

    @override
    def create_plugin(self, trainer_config):
        from lightning.pytorch.plugins.precision.fsdp import FSDPPrecision

        return FSDPPrecision(precision=self.precision)


@plugin_registry.register
class HalfPrecisionPluginConfig(PluginConfigBase):
    name: Literal["half_precision"] = "half_precision"

    precision: Literal["bf16-true", "16-true"]
    """Whether to use ``torch.float16`` (``'16-true'``) or ``torch.bfloat16`` (``'bf16-true'``)."""

    @override
    def create_plugin(self, trainer_config):
        from lightning.pytorch.plugins.precision.half import HalfPrecision

        return HalfPrecision(precision=self.precision)


@plugin_registry.register
class TransformerEnginePluginConfig(PluginConfigBase):
    name: Literal["transformer_engine_precision"] = "transformer_engine_precision"

    weights_dtype: DTypeConfig
    """The weights dtype to use."""

    recipe: dict[str, Any] | None = None
    """Recipe for the DelayedScaling configuration in dict format."""

    replace_layers: bool | None = None
    """Whether to replace ``Linear`` and ``LayerNorm`` layers automatically with their
    Transformer Engine alternatives."""

    fallback_compute_dtype: DTypeConfig | None = None
    """The compute dtype to use for operations that don't support fp8 autocast.
    Defaults to the same as weights_dtype."""

    @override
    def create_plugin(self, trainer_config):
        from lightning.pytorch.plugins.precision.transformer_engine import (
            TransformerEnginePrecision,
        )

        return TransformerEnginePrecision(
            weights_dtype=self.weights_dtype.torch_dtype,
            recipe=self.recipe,
            replace_layers=self.replace_layers,
            fallback_compute_dtype=self.fallback_compute_dtype.torch_dtype
            if self.fallback_compute_dtype
            else None,
        )


@plugin_registry.register
class XLAPluginConfig(PluginConfigBase):
    name: Literal["xla_precision"] = "xla_precision"

    precision: Literal["32-true", "16-true", "bf16-true"]
    """Full precision (32-true) or half precision (16-true, bf16-true)."""

    @override
    def create_plugin(self, trainer_config):
        from lightning.pytorch.plugins.precision.xla import XLAPrecision

        return XLAPrecision(precision=self.precision)



from __future__ import annotations

__codegen__ = True

from nshtrainer.nn.nonlinearity import ELUNonlinearityConfig as ELUNonlinearityConfig
from nshtrainer.nn.nonlinearity import GELUNonlinearityConfig as GELUNonlinearityConfig
from nshtrainer.nn.nonlinearity import (
    LeakyReLUNonlinearityConfig as LeakyReLUNonlinearityConfig,
)
from nshtrainer.nn.nonlinearity import MishNonlinearityConfig as MishNonlinearityConfig
from nshtrainer.nn.nonlinearity import NonlinearityConfig as NonlinearityConfig
from nshtrainer.nn.nonlinearity import NonlinearityConfigBase as NonlinearityConfigBase
from nshtrainer.nn.nonlinearity import PReLUConfig as PReLUConfig
from nshtrainer.nn.nonlinearity import ReLUNonlinearityConfig as ReLUNonlinearityConfig
from nshtrainer.nn.nonlinearity import (
    SigmoidNonlinearityConfig as SigmoidNonlinearityConfig,
)
from nshtrainer.nn.nonlinearity import SiLUNonlinearityConfig as SiLUNonlinearityConfig
from nshtrainer.nn.nonlinearity import (
    SoftmaxNonlinearityConfig as SoftmaxNonlinearityConfig,
)
from nshtrainer.nn.nonlinearity import (
    SoftplusNonlinearityConfig as SoftplusNonlinearityConfig,
)
from nshtrainer.nn.nonlinearity import (
    SoftsignNonlinearityConfig as SoftsignNonlinearityConfig,
)
from nshtrainer.nn.nonlinearity import (
    SwiGLUNonlinearityConfig as SwiGLUNonlinearityConfig,
)
from nshtrainer.nn.nonlinearity import (
    SwishNonlinearityConfig as SwishNonlinearityConfig,
)
from nshtrainer.nn.nonlinearity import TanhNonlinearityConfig as TanhNonlinearityConfig
from nshtrainer.nn.nonlinearity import nonlinearity_registry as nonlinearity_registry

__all__ = [
    "ELUNonlinearityConfig",
    "GELUNonlinearityConfig",
    "LeakyReLUNonlinearityConfig",
    "MishNonlinearityConfig",
    "NonlinearityConfig",
    "NonlinearityConfigBase",
    "PReLUConfig",
    "ReLUNonlinearityConfig",
    "SiLUNonlinearityConfig",
    "SigmoidNonlinearityConfig",
    "SoftmaxNonlinearityConfig",
    "SoftplusNonlinearityConfig",
    "SoftsignNonlinearityConfig",
    "SwiGLUNonlinearityConfig",
    "SwishNonlinearityConfig",
    "TanhNonlinearityConfig",
    "nonlinearity_registry",
]



from __future__ import annotations

from . import duration as duration
from .dtype import DTypeConfig as DTypeConfig
from .duration import DurationConfig as DurationConfig
from .duration import EpochsConfig as EpochsConfig
from .duration import StepsConfig as StepsConfig



from __future__ import annotations

import logging
import math
from typing import Any, Literal

from lightning.pytorch import LightningModule, Trainer
from lightning.pytorch.callbacks import Callback
from typing_extensions import final, override

from .base import CallbackConfigBase, callback_registry

log = logging.getLogger(__name__)


@final
@callback_registry.register
class LogEpochCallbackConfig(CallbackConfigBase):
    name: Literal["log_epoch"] = "log_epoch"

    @override
    def create_callbacks(self, trainer_config):
        yield LogEpochCallback()


class LogEpochCallback(Callback):
    def __init__(self, metric_name: str = "computed_epoch"):
        super().__init__()

        self.metric_name = metric_name

    @override
    def on_train_batch_start(
        self, trainer: Trainer, pl_module: LightningModule, batch: Any, batch_idx: int
    ):
        if trainer.logger is None:
            return

        # If trainer.num_training_batches is not set or is nan/inf, we cannot calculate the epoch
        if (
            not trainer.num_training_batches
            or math.isnan(trainer.num_training_batches)
            or math.isinf(trainer.num_training_batches)
        ):
            log.warning("Trainer has no valid num_training_batches. Cannot log epoch.")
            return

        epoch = pl_module.global_step / trainer.num_training_batches
        pl_module.log(self.metric_name, epoch, on_step=True, on_epoch=False)



from __future__ import annotations

__codegen__ = True

from nshtrainer.callbacks.finite_checks import CallbackConfigBase as CallbackConfigBase
from nshtrainer.callbacks.finite_checks import (
    FiniteChecksCallbackConfig as FiniteChecksCallbackConfig,
)
from nshtrainer.callbacks.finite_checks import callback_registry as callback_registry

__all__ = [
    "CallbackConfigBase",
    "FiniteChecksCallbackConfig",
    "callback_registry",
]



from __future__ import annotations

import logging
from dataclasses import dataclass
from pathlib import Path

log = logging.getLogger(__name__)


@dataclass
class DistributedPredictionResult:
    """Represents the results of a distributed prediction run.

    This dataclass provides easy access to both raw and processed prediction data.
    """

    root_dir: Path
    """Root directory where predictions are stored."""

    @property
    def raw_dir(self) -> Path:
        """Directory containing raw prediction data."""
        return self.root_dir / "raw"

    @property
    def processed_dir(self) -> Path:
        """Directory containing processed prediction data."""
        return self.root_dir / "processed"

    def get_raw_predictions(self, dataloader_idx: int = 0) -> Path:
        """Get the directory containing raw predictions for a specific dataloader.

        Args:
            dataloader_idx: Index of the dataloader

        Returns:
            Path to the raw predictions directory for the specified dataloader
        """
        raw_loader_dir = self.raw_dir / f"dataloader_{dataloader_idx}"
        if not raw_loader_dir.exists():
            log.warning(f"Raw predictions directory {raw_loader_dir} does not exist.")
        return raw_loader_dir

    def get_processed_reader(self, dataloader_idx: int = 0):
        """Get a reader for processed predictions from a specific dataloader.

        Args:
            dataloader_idx: Index of the dataloader

        Returns:
            A DistributedPredictionReader for the processed predictions, or None if no data exists
        """
        from ..callbacks.distributed_prediction_writer import (
            DistributedPredictionReader,
        )

        processed_loader_dir = self.processed_dir / f"dataloader_{dataloader_idx}"
        if not processed_loader_dir.exists():
            log.warning(
                f"Processed predictions directory {processed_loader_dir} does not exist."
            )
            return None

        return DistributedPredictionReader(processed_loader_dir)

    @classmethod
    def load(cls, path: Path | str):
        """Load prediction results from a directory.

        Args:
            path: Path to the predictions directory

        Returns:
            A DistributedPredictionResult instance
        """
        path = Path(path)
        if not path.exists():
            raise FileNotFoundError(f"Predictions directory {path} does not exist.")

        return cls(root_dir=path)



from __future__ import annotations

import json
import logging
from typing import Any, cast

import nshconfig as C
from lightning.pytorch import LightningDataModule, Trainer

log = logging.getLogger(__name__)


def _dict(obj: Any):
    if isinstance(obj, C.Config):
        return obj.model_dump(mode="json")

    try:
        return dict(obj)
    except Exception:
        return json.loads(
            json.dumps(obj, default=lambda o: str(o), indent=4, sort_keys=True)
        )


def _dict_and_clean(obj: Any):
    d = _dict(obj)

    # Remove LightningCLI's internal hparam
    d = {k: v for k, v in d.items() if k != "_class_path"}
    return d


def _log_hyperparams(trainer: Trainer) -> None:
    if not trainer.loggers:
        return

    hparams_to_log: dict[str, Any] = {}

    from .trainer import Trainer

    if isinstance(trainer, Trainer):
        hparams_to_log["trainer"] = _dict_and_clean(trainer.hparams)

    if (
        pl_module := trainer.lightning_module
    ) is not None and pl_module._log_hyperparams:
        hparams_to_log["model"] = _dict_and_clean(pl_module.hparams_initial)

    if (
        datamodule := cast(
            LightningDataModule | None, getattr(trainer, "datamodule", None)
        )
    ) is not None and (datamodule._log_hyperparams):
        hparams_to_log["datamodule"] = _dict_and_clean(datamodule.hparams_initial)

    for logger in trainer.loggers:
        logger.log_hyperparams(hparams_to_log)
        logger.log_graph(pl_module)
        logger.save()


def patch_log_hparams_function():
    try:
        import lightning.pytorch.loggers.utilities
        import lightning.pytorch.trainer.trainer

        lightning.pytorch.loggers.utilities._log_hyperparams = _log_hyperparams
        lightning.pytorch.trainer.trainer._log_hyperparams = _log_hyperparams
        log.info(
            "Patched lightning.pytorch's _log_hyperparams to use nshtrainer's version"
        )
    except ImportError:
        pass

    try:
        import pytorch_lightning.loggers.utilities
        import pytorch_lightning.trainer.trainer

        pytorch_lightning.loggers.utilities._log_hyperparams = _log_hyperparams
        pytorch_lightning.trainer.trainer._log_hyperparams = _log_hyperparams
        log.info(
            "Patched pytorch_lightning's _log_hyperparams to use nshtrainer's version"
        )
    except ImportError:
        pass



from __future__ import annotations

import datetime
import logging
import time
from dataclasses import dataclass
from typing import Any, Literal

from lightning.pytorch.callbacks.callback import Callback
from typing_extensions import TypeAliasType, override

log = logging.getLogger(__name__)

Stage = TypeAliasType("Stage", Literal["train", "validate", "test", "predict"])
ALL_STAGES = ("train", "validate", "test", "predict")


@dataclass
class TimeInfo:
    datetime: datetime.datetime
    monotonic: float


class RuntimeTrackerCallback(Callback):
    def __init__(self):
        super().__init__()
        self._start_time: dict[Stage, TimeInfo] = {}
        self._end_time: dict[Stage, TimeInfo] = {}
        self._offsets = {stage: datetime.timedelta() for stage in ALL_STAGES}

    def start_time(self, stage: Stage) -> TimeInfo | None:
        """Return the start time of a particular stage"""
        return self._start_time.get(stage)

    def end_time(self, stage: Stage) -> TimeInfo | None:
        """Return the end time of a particular stage"""
        return self._end_time.get(stage)

    def time_elapsed(self, stage: Stage) -> datetime.timedelta:
        """Return the time elapsed for a particular stage"""
        start = self.start_time(stage)
        end = self.end_time(stage)
        offset = self._offsets[stage]
        if start is None:
            return offset
        if end is None:
            current = TimeInfo(datetime.datetime.now(), time.monotonic())
            return (
                datetime.timedelta(seconds=current.monotonic - start.monotonic) + offset
            )
        return datetime.timedelta(seconds=end.monotonic - start.monotonic) + offset

    def _record_time(self, stage: Stage, time_dict: dict[Stage, TimeInfo]):
        time_dict[stage] = TimeInfo(datetime.datetime.now(), time.monotonic())

    @override
    def on_train_start(self, trainer, pl_module):
        self._record_time("train", self._start_time)

    @override
    def on_train_end(self, trainer, pl_module):
        self._record_time("train", self._end_time)

    @override
    def on_validation_start(self, trainer, pl_module):
        self._record_time("validate", self._start_time)

    @override
    def on_validation_end(self, trainer, pl_module):
        self._record_time("validate", self._end_time)

    @override
    def on_test_start(self, trainer, pl_module):
        self._record_time("test", self._start_time)

    @override
    def on_test_end(self, trainer, pl_module):
        self._record_time("test", self._end_time)

    @override
    def on_predict_start(self, trainer, pl_module):
        self._record_time("predict", self._start_time)

    @override
    def on_predict_end(self, trainer, pl_module):
        self._record_time("predict", self._end_time)

    @override
    def state_dict(self) -> dict[str, Any]:
        return {
            "time_elapsed": {
                stage: self.time_elapsed(stage).total_seconds() for stage in ALL_STAGES
            },
            "start_times": {
                stage: (info.datetime.isoformat(), info.monotonic)
                for stage, info in self._start_time.items()
            },
            "end_times": {
                stage: (info.datetime.isoformat(), info.monotonic)
                for stage, info in self._end_time.items()
            },
        }

    @override
    def load_state_dict(self, state_dict: dict[str, Any]):
        time_elapsed: dict[Stage, float] = state_dict.get("time_elapsed", {})
        for stage in ALL_STAGES:
            self._offsets[stage] = datetime.timedelta(
                seconds=time_elapsed.get(stage, 0)
            )

        start_times: dict[Stage, tuple[str, float]] = state_dict.get("start_times", {})
        for stage, (dt_str, monotonic) in start_times.items():
            self._start_time[stage] = TimeInfo(
                datetime.datetime.fromisoformat(dt_str), monotonic
            )

        end_times: dict[Stage, tuple[str, float]] = state_dict.get("end_times", {})
        for stage, (dt_str, monotonic) in end_times.items():
            self._end_time[stage] = TimeInfo(
                datetime.datetime.fromisoformat(dt_str), monotonic
            )



from __future__ import annotations

__codegen__ = True

from nshtrainer.loggers.wandb import CallbackConfigBase as CallbackConfigBase
from nshtrainer.loggers.wandb import LoggerConfigBase as LoggerConfigBase
from nshtrainer.loggers.wandb import WandbLoggerConfig as WandbLoggerConfig
from nshtrainer.loggers.wandb import (
    WandbUploadCodeCallbackConfig as WandbUploadCodeCallbackConfig,
)
from nshtrainer.loggers.wandb import (
    WandbWatchCallbackConfig as WandbWatchCallbackConfig,
)
from nshtrainer.loggers.wandb import logger_registry as logger_registry

__all__ = [
    "CallbackConfigBase",
    "LoggerConfigBase",
    "WandbLoggerConfig",
    "WandbUploadCodeCallbackConfig",
    "WandbWatchCallbackConfig",
    "logger_registry",
]



from __future__ import annotations

import logging
from typing import Any, Literal

from typing_extensions import override

from ._base import BaseProfilerConfig

log = logging.getLogger(__name__)


class PyTorchProfilerConfig(BaseProfilerConfig):
    name: Literal["pytorch"] = "pytorch"

    group_by_input_shapes: bool = False
    """Include operator input shapes and group calls by shape."""

    emit_nvtx: bool = False
    """
    Context manager that makes every autograd operation emit an NVTX range
        Run::

            nvprof --profile-from-start off -o trace_name.prof -- <regular command here>

        To visualize, you can either use::

            nvvp trace_name.prof
            torch.autograd.profiler.load_nvprof(path)
    """

    export_to_chrome: bool = True
    """
    Whether to export the sequence of profiled operators for Chrome.
        It will generate a ``.json`` file which can be read by Chrome.
    """

    row_limit: int = 20
    """
    Limit the number of rows in a table, ``-1`` is a special value that
        removes the limit completely.
    """

    sort_by_key: str | None = None
    """
    Attribute used to sort entries. By default
        they are printed in the same order as they were registered.
        Valid keys include: ``cpu_time``, ``cuda_time``, ``cpu_time_total``,
        ``cuda_time_total``, ``cpu_memory_usage``, ``cuda_memory_usage``,
        ``self_cpu_memory_usage``, ``self_cuda_memory_usage``, ``count``.
    """

    record_module_names: bool = True
    """Whether to add module names while recording autograd operation."""

    table_kwargs: dict[str, Any] | None = None
    """Dictionary with keyword arguments for the summary table."""

    additional_profiler_kwargs: dict[str, Any] = {}
    """Keyword arguments for the PyTorch profiler. This depends on your PyTorch version"""

    @override
    def create_profiler(self, trainer_config):
        from lightning.pytorch.profilers.pytorch import PyTorchProfiler

        if (dirpath := self.dirpath) is None:
            dirpath = trainer_config.directory.resolve_subdirectory(
                trainer_config.id, "profile"
            )

        if (filename := self.filename) is None:
            filename = f"{trainer_config.id}_profile.txt"

        return PyTorchProfiler(
            group_by_input_shapes=self.group_by_input_shapes,
            emit_nvtx=self.emit_nvtx,
            export_to_chrome=self.export_to_chrome,
            row_limit=self.row_limit,
            sort_by_key=self.sort_by_key,
            record_module_names=self.record_module_names,
            table_kwargs=self.table_kwargs,
            dirpath=dirpath,
            filename=filename,
            **self.additional_profiler_kwargs,
        )



from __future__ import annotations

from typing import Literal

from lightning.pytorch.callbacks import LearningRateMonitor
from typing_extensions import final

from .base import CallbackConfigBase, callback_registry


@final
@callback_registry.register
class LearningRateMonitorConfig(CallbackConfigBase):
    name: Literal["learning_rate_monitor"] = "learning_rate_monitor"

    logging_interval: Literal["step", "epoch"] | None = None
    """
    Set to 'epoch' or 'step' to log 'lr' of all optimizers at the same interval, set to None to log at individual interval according to the 'interval' key of each scheduler. Defaults to None.
    """

    log_momentum: bool = False
    """
    Option to also log the momentum values of the optimizer, if the optimizer has the 'momentum' or 'betas' attribute. Defaults to False.
    """

    log_weight_decay: bool = False
    """
    Option to also log the weight decay values of the optimizer. Defaults to False.
    """

    def create_callbacks(self, trainer_config):
        yield LearningRateMonitor(
            logging_interval=self.logging_interval,
            log_momentum=self.log_momentum,
            log_weight_decay=self.log_weight_decay,
        )



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterable
from typing import Annotated, Any, Literal, Tuple, Union

import nshconfig as C
import torch.nn as nn
from torch import Tensor
from torch.optim import Optimizer
from typing_extensions import TypeAliasType, final, override


class OptimizerConfigBase(C.Config, ABC):
    @abstractmethod
    def create_optimizer(
        self,
        parameters: Iterable[nn.Parameter] | Iterable[dict[str, Any]],
    ) -> Optimizer: ...


optimizer_registry = C.Registry(OptimizerConfigBase, discriminator="name")


@final
@optimizer_registry.register
class AdamWConfig(OptimizerConfigBase):
    name: Literal["adamw"] = "adamw"

    lr: float
    """Learning rate for the optimizer."""

    weight_decay: float = 1.0e-2
    """Weight decay (L2 penalty) for the optimizer."""

    betas: tuple[float, float] = (0.9, 0.999)
    """
    Betas for the optimizer:
    (beta1, beta2) are the coefficients used for computing running averages of
    gradient and its square.
    """

    eps: float = 1e-8
    """Term added to the denominator to improve numerical stability."""

    amsgrad: bool = False
    """Whether to use the AMSGrad variant of this algorithm."""

    maximize: bool = False
    """Maximize the objective with respect to the params, instead of minimizing."""

    foreach: bool | None = None
    """Whether foreach implementation of optimizer is used."""

    capturable: bool = False
    """Whether this instance is safe to capture in a CUDA graph."""

    differentiable: bool = False
    """Whether autograd should occur through the optimizer step in training."""

    @override
    def create_optimizer(
        self,
        parameters: Iterable[nn.Parameter] | Iterable[dict[str, Any]],
    ):
        from torch.optim import AdamW

        return AdamW(
            parameters,
            lr=self.lr,
            weight_decay=self.weight_decay,
            betas=self.betas,
            eps=self.eps,
            amsgrad=self.amsgrad,
            maximize=self.maximize,
            foreach=self.foreach,
            capturable=self.capturable,
            differentiable=self.differentiable,
        )


@final
@optimizer_registry.register
class AdafactorConfig(OptimizerConfigBase):
    name: Literal["adafactor"] = "adafactor"
    lr: float
    """Learning rate for the optimizer. If None, uses relative step size."""

    eps1: float | None = None
    """Term added to the denominator to improve numerical stability (default: None)."""

    eps2: float = 1e-3
    """Term added to the denominator to improve numerical stability (default: 1e-3)."""

    beta2_decay: float = -0.8
    """Coefficient used for computing running averages of square gradient (default: -0.8)."""

    weight_decay: float = 0.0
    """Weight decay (L2 penalty) (default: 0.0)."""

    maximize: bool = False
    """Maximize the params based on the objective, instead of minimizing."""

    @override
    def create_optimizer(
        self,
        parameters: Iterable[nn.Parameter] | Iterable[dict[str, Any]],
    ):
        from torch.optim import Adafactor

        return Adafactor(
            parameters,
            lr=self.lr,
            eps=(self.eps1, self.eps2),
            beta2_decay=self.beta2_decay,
            weight_decay=self.weight_decay,
            maximize=self.maximize,
        )


@final
@optimizer_registry.register
class AdadeltaConfig(OptimizerConfigBase):
    name: Literal["adadelta"] = "adadelta"

    lr: float
    """Learning rate for the optimizer."""

    rho: float = 0.9
    """Coefficient used for computing a running average of squared gradients."""

    eps: float = 1e-6
    """Term added to the denominator to improve numerical stability."""

    weight_decay: float = 0.0
    """Weight decay (L2 penalty) for the optimizer."""

    maximize: bool = False
    """Maximize the params based on the objective, instead of minimizing."""

    foreach: bool | None = None
    """Whether foreach implementation of optimizer is used."""

    capturable: bool = False
    """Whether this instance is safe to capture in a CUDA graph."""

    differentiable: bool = False
    """Whether autograd should occur through the optimizer step in training."""

    @override
    def create_optimizer(
        self,
        parameters: Iterable[nn.Parameter] | Iterable[dict[str, Any]],
    ):
        from torch.optim import Adadelta

        return Adadelta(
            parameters,
            lr=self.lr,
            rho=self.rho,
            eps=self.eps,
            weight_decay=self.weight_decay,
            maximize=self.maximize,
            foreach=self.foreach,
            capturable=self.capturable,
            differentiable=self.differentiable,
        )


@final
@optimizer_registry.register
class AdagradConfig(OptimizerConfigBase):
    name: Literal["adagrad"] = "adagrad"

    lr: float
    """Learning rate for the optimizer."""

    lr_decay: float = 0.0
    """Learning rate decay."""

    weight_decay: float = 0.0
    """Weight decay (L2 penalty) for the optimizer."""

    initial_accumulator_value: float = 0.0
    """Initial value for the accumulator."""

    eps: float = 1e-10
    """Term added to the denominator to improve numerical stability."""

    maximize: bool = False
    """Maximize the params based on the objective, instead of minimizing."""

    foreach: bool | None = None
    """Whether foreach implementation of optimizer is used."""

    differentiable: bool = False
    """Whether autograd should occur through the optimizer step in training."""

    fused: bool | None = None
    """Whether the fused implementation is used."""

    @override
    def create_optimizer(
        self,
        parameters: Iterable[nn.Parameter] | Iterable[dict[str, Any]],
    ):
        from torch.optim import Adagrad

        return Adagrad(
            parameters,
            lr=self.lr,
            lr_decay=self.lr_decay,
            weight_decay=self.weight_decay,
            initial_accumulator_value=self.initial_accumulator_value,
            eps=self.eps,
            maximize=self.maximize,
            foreach=self.foreach,
            differentiable=self.differentiable,
            fused=self.fused,
        )


@final
@optimizer_registry.register
class AdamConfig(OptimizerConfigBase):
    name: Literal["adam"] = "adam"

    lr: float
    """Learning rate for the optimizer."""

    betas: tuple[float, float] = (0.9, 0.999)
    """Coefficients used for computing running averages of gradient and its square."""

    eps: float = 1e-8
    """Term added to the denominator to improve numerical stability."""

    weight_decay: float = 0.0
    """Weight decay (L2 penalty) for the optimizer."""

    amsgrad: bool = False
    """Whether to use the AMSGrad variant of this algorithm."""

    maximize: bool = False
    """Maximize the params based on the objective, instead of minimizing."""

    foreach: bool | None = None
    """Whether foreach implementation of optimizer is used."""

    capturable: bool = False
    """Whether this instance is safe to capture in a CUDA graph."""

    differentiable: bool = False
    """Whether autograd should occur through the optimizer step in training."""

    fused: bool | None = None
    """Whether the fused implementation is used."""

    @override
    def create_optimizer(
        self,
        parameters: Iterable[nn.Parameter] | Iterable[dict[str, Any]],
    ):
        from torch.optim import Adam

        return Adam(
            parameters,
            lr=self.lr,
            betas=self.betas,
            eps=self.eps,
            weight_decay=self.weight_decay,
            amsgrad=self.amsgrad,
            maximize=self.maximize,
            foreach=self.foreach,
            capturable=self.capturable,
            differentiable=self.differentiable,
            fused=self.fused,
        )


@final
@optimizer_registry.register
class AdamaxConfig(OptimizerConfigBase):
    name: Literal["adamax"] = "adamax"

    lr: float
    """Learning rate for the optimizer."""

    betas: tuple[float, float] = (0.9, 0.999)
    """Coefficients used for computing running averages of gradient and its square."""

    eps: float = 1e-8
    """Term added to the denominator to improve numerical stability."""

    weight_decay: float = 0.0
    """Weight decay (L2 penalty) for the optimizer."""

    maximize: bool = False
    """Maximize the params based on the objective, instead of minimizing."""

    foreach: bool | None = None
    """Whether foreach implementation of optimizer is used."""

    capturable: bool = False
    """Whether this instance is safe to capture in a CUDA graph."""

    differentiable: bool = False
    """Whether autograd should occur through the optimizer step in training."""

    @override
    def create_optimizer(
        self,
        parameters: Iterable[nn.Parameter] | Iterable[dict[str, Any]],
    ):
        from torch.optim import Adamax

        return Adamax(
            parameters,
            lr=self.lr,
            betas=self.betas,
            eps=self.eps,
            weight_decay=self.weight_decay,
            maximize=self.maximize,
            foreach=self.foreach,
            capturable=self.capturable,
            differentiable=self.differentiable,
        )


@final
@optimizer_registry.register
class ASGDConfig(OptimizerConfigBase):
    name: Literal["asgd"] = "asgd"

    lr: float
    """Learning rate for the optimizer."""

    lambd: float = 1e-4
    """Decay term."""

    alpha: float = 0.75
    """Power for eta update."""

    t0: float = 1e6
    """Point at which to start averaging."""

    weight_decay: float = 0.0
    """Weight decay (L2 penalty) for the optimizer."""

    maximize: bool = False
    """Maximize the params based on the objective, instead of minimizing."""

    @override
    def create_optimizer(
        self,
        parameters: Iterable[nn.Parameter] | Iterable[dict[str, Any]],
    ):
        from torch.optim import ASGD

        return ASGD(
            parameters,
            lr=self.lr,
            lambd=self.lambd,
            alpha=self.alpha,
            t0=self.t0,
            weight_decay=self.weight_decay,
            maximize=self.maximize,
        )


@final
@optimizer_registry.register
class NAdamConfig(OptimizerConfigBase):
    name: Literal["nadam"] = "nadam"

    lr: float
    """Learning rate for the optimizer."""

    betas: tuple[float, float] = (0.9, 0.999)
    """Coefficients used for computing running averages of gradient and its square."""

    eps: float = 1e-8
    """Term added to the denominator to improve numerical stability."""

    weight_decay: float = 0.0
    """Weight decay (L2 penalty) for the optimizer."""

    momentum_decay: float = 4e-3
    """Momentum decay."""

    decoupled_weight_decay: bool = False
    """Whether to use decoupled weight decay."""

    maximize: bool = False
    """Maximize the params based on the objective, instead of minimizing."""

    foreach: bool | None = None
    """Whether foreach implementation of optimizer is used."""

    capturable: bool = False
    """Whether this instance is safe to capture in a CUDA graph."""

    differentiable: bool = False
    """Whether autograd should occur through the optimizer step in training."""

    @override
    def create_optimizer(
        self,
        parameters: Iterable[nn.Parameter] | Iterable[dict[str, Any]],
    ):
        from torch.optim import NAdam

        return NAdam(
            parameters,
            lr=self.lr,
            betas=self.betas,
            eps=self.eps,
            weight_decay=self.weight_decay,
            momentum_decay=self.momentum_decay,
            decoupled_weight_decay=self.decoupled_weight_decay,
            maximize=self.maximize,
            foreach=self.foreach,
            capturable=self.capturable,
            differentiable=self.differentiable,
        )


@final
@optimizer_registry.register
class RAdamConfig(OptimizerConfigBase):
    name: Literal["radam"] = "radam"

    lr: float
    """Learning rate for the optimizer."""

    betas: tuple[float, float] = (0.9, 0.999)
    """Coefficients used for computing running averages of gradient and its square."""

    eps: float = 1e-8
    """Term added to the denominator to improve numerical stability."""

    weight_decay: float = 0.0
    """Weight decay (L2 penalty) for the optimizer."""

    decoupled_weight_decay: bool = False
    """Whether to use decoupled weight decay."""

    maximize: bool = False
    """Maximize the params based on the objective, instead of minimizing."""

    foreach: bool | None = None
    """Whether foreach implementation of optimizer is used."""

    capturable: bool = False
    """Whether this instance is safe to capture in a CUDA graph."""

    differentiable: bool = False
    """Whether autograd should occur through the optimizer step in training."""

    @override
    def create_optimizer(
        self,
        parameters: Iterable[nn.Parameter] | Iterable[dict[str, Any]],
    ):
        from torch.optim import RAdam

        return RAdam(
            parameters,
            lr=self.lr,
            betas=self.betas,
            eps=self.eps,
            weight_decay=self.weight_decay,
            decoupled_weight_decay=self.decoupled_weight_decay,
            maximize=self.maximize,
            foreach=self.foreach,
            capturable=self.capturable,
            differentiable=self.differentiable,
        )


@final
@optimizer_registry.register
class RMSpropConfig(OptimizerConfigBase):
    name: Literal["rmsprop"] = "rmsprop"

    lr: float
    """Learning rate for the optimizer."""

    alpha: float = 0.99
    """Smoothing constant."""

    eps: float = 1e-8
    """Term added to the denominator to improve numerical stability."""

    weight_decay: float = 0.0
    """Weight decay (L2 penalty) for the optimizer."""

    momentum: float = 0.0
    """Momentum factor."""

    centered: bool = False
    """If True, compute the centered RMSProp, the gradient is normalized by an estimation of its variance."""

    maximize: bool = False
    """Maximize the params based on the objective, instead of minimizing."""

    foreach: bool | None = None
    """Whether foreach implementation of optimizer is used."""

    capturable: bool = False
    """Whether this instance is safe to capture in a CUDA graph."""

    differentiable: bool = False
    """Whether autograd should occur through the optimizer step in training."""

    @override
    def create_optimizer(
        self,
        parameters: Iterable[nn.Parameter] | Iterable[dict[str, Any]],
    ):
        from torch.optim import RMSprop

        return RMSprop(
            parameters,
            lr=self.lr,
            alpha=self.alpha,
            eps=self.eps,
            weight_decay=self.weight_decay,
            momentum=self.momentum,
            centered=self.centered,
            maximize=self.maximize,
            foreach=self.foreach,
            capturable=self.capturable,
            differentiable=self.differentiable,
        )


@final
@optimizer_registry.register
class RpropConfig(OptimizerConfigBase):
    name: Literal["rprop"] = "rprop"

    lr: float
    """Learning rate for the optimizer."""

    etas: tuple[float, float] = (0.5, 1.2)
    """Pair of (etaminus, etaplus), multiplicative increase and decrease factors."""

    step_sizes: tuple[float, float] = (1e-6, 50.0)
    """Pair of minimal and maximal allowed step sizes."""

    maximize: bool = False
    """Maximize the params based on the objective, instead of minimizing."""

    @override
    def create_optimizer(
        self,
        parameters: Iterable[nn.Parameter] | Iterable[dict[str, Any]],
    ):
        from torch.optim import Rprop

        return Rprop(
            parameters,
            lr=self.lr,
            etas=self.etas,
            step_sizes=self.step_sizes,
            maximize=self.maximize,
        )


@final
@optimizer_registry.register
class SGDConfig(OptimizerConfigBase):
    name: Literal["sgd"] = "sgd"

    lr: float
    """Learning rate for the optimizer."""

    momentum: float = 0.0
    """Momentum factor."""

    dampening: float = 0.0
    """Dampening for momentum."""

    weight_decay: float = 0.0
    """Weight decay (L2 penalty) for the optimizer."""

    nesterov: bool = False
    """Enables Nesterov momentum."""

    maximize: bool = False
    """Maximize the params based on the objective, instead of minimizing."""

    foreach: bool | None = None
    """Whether foreach implementation of optimizer is used."""

    differentiable: bool = False
    """Whether autograd should occur through the optimizer step in training."""

    fused: bool | None = None
    """Whether the fused implementation is used."""

    @override
    def create_optimizer(
        self,
        parameters: Iterable[nn.Parameter] | Iterable[dict[str, Any]],
    ):
        from torch.optim import SGD

        return SGD(
            parameters,
            lr=self.lr,
            momentum=self.momentum,
            dampening=self.dampening,
            weight_decay=self.weight_decay,
            nesterov=self.nesterov,
            maximize=self.maximize,
            foreach=self.foreach,
            differentiable=self.differentiable,
            fused=self.fused,
        )


OptimizerConfig = TypeAliasType(
    "OptimizerConfig",
    Annotated[OptimizerConfigBase, optimizer_registry.DynamicResolution()],
)



from __future__ import annotations

from ..callbacks import callback_registry as callback_registry
from ..callbacks.distributed_prediction_writer import (
    DistributedPredictionReader as DistributedPredictionReader,
)
from ._config import TrainerConfig as TrainerConfig
from ._distributed_prediction_result import (
    DistributedPredictionResult as DistributedPredictionResult,
)
from .accelerator import accelerator_registry as accelerator_registry
from .plugin import plugin_registry as plugin_registry
from .trainer import Trainer as Trainer



from __future__ import annotations

import getpass
import importlib.metadata
import inspect
import logging
import os
import platform
import socket
import sys
from datetime import timedelta
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast

import nshconfig as C
import psutil
import torch
from lightning.pytorch import LightningDataModule, LightningModule
from packaging import version
from typing_extensions import Self

from .slurm import parse_slurm_node_list

if TYPE_CHECKING:
    from ..trainer._config import TrainerConfig


log = logging.getLogger(__name__)


class EnvironmentClassInformationConfig(C.Config):
    """Configuration for class information in the environment."""

    name: str | None = None
    """The name of the class."""

    module: str | None = None
    """The module where the class is defined."""

    full_name: str | None = None
    """The fully qualified name of the class."""

    file_path: Path | None = None
    """The file path where the class is defined."""

    source_file_path: Path | None = None
    """The source file path of the class, if available."""

    @classmethod
    def empty(cls):
        return cls(
            name=None,
            module=None,
            full_name=None,
            file_path=None,
            source_file_path=None,
        )

    @classmethod
    def from_class(cls, cls_: type):
        name = cls_.__name__
        module = cls_.__module__
        full_name = f"{cls_.__module__}.{cls_.__qualname__}"

        file_path = inspect.getfile(cls_)
        source_file_path = inspect.getsourcefile(cls_)
        return cls(
            name=name,
            module=module,
            full_name=full_name,
            file_path=Path(file_path),
            source_file_path=Path(source_file_path) if source_file_path else None,
        )

    @classmethod
    def from_instance(cls, instance: object):
        return cls.from_class(type(instance))


class EnvironmentSLURMInformationConfig(C.Config):
    """Configuration for SLURM environment information."""

    hostname: str | None = None
    """The hostname of the current node."""

    hostnames: list[str] | None = None
    """List of hostnames for all nodes in the job."""

    job_id: str | None = None
    """The SLURM job ID."""

    raw_job_id: str | None = None
    """The raw SLURM job ID."""

    array_job_id: str | None = None
    """The SLURM array job ID, if applicable."""

    array_task_id: str | None = None
    """The SLURM array task ID, if applicable."""

    num_tasks: int | None = None
    """The number of tasks in the SLURM job."""

    num_nodes: int | None = None
    """The number of nodes in the SLURM job."""

    node: str | int | None = None
    """The node ID or name."""

    global_rank: int | None = None
    """The global rank of the current process."""

    local_rank: int | None = None
    """The local rank of the current process within its node."""

    @classmethod
    def empty(cls):
        return cls(
            hostname=None,
            hostnames=None,
            job_id=None,
            raw_job_id=None,
            array_job_id=None,
            array_task_id=None,
            num_tasks=None,
            num_nodes=None,
            node=None,
            global_rank=None,
            local_rank=None,
        )

    @classmethod
    def from_current_environment(cls):
        try:
            from lightning.fabric.plugins.environments.slurm import SLURMEnvironment

            if not SLURMEnvironment.detect():
                return None

            hostname = socket.gethostname()
            hostnames = [hostname]
            if node_list := os.environ.get("SLURM_JOB_NODELIST", ""):
                hostnames = parse_slurm_node_list(node_list)

            raw_job_id = os.environ["SLURM_JOB_ID"]
            job_id = raw_job_id
            array_job_id = os.environ.get("SLURM_ARRAY_JOB_ID")
            array_task_id = os.environ.get("SLURM_ARRAY_TASK_ID")
            if array_job_id and array_task_id:
                job_id = f"{array_job_id}_{array_task_id}"

            num_tasks = int(os.environ["SLURM_NTASKS"])
            num_nodes = int(os.environ["SLURM_JOB_NUM_NODES"])

            node_id = os.environ.get("SLURM_NODEID")

            global_rank = int(os.environ["SLURM_PROCID"])
            local_rank = int(os.environ["SLURM_LOCALID"])

            return cls(
                hostname=hostname,
                hostnames=hostnames,
                job_id=job_id,
                raw_job_id=raw_job_id,
                array_job_id=array_job_id,
                array_task_id=array_task_id,
                num_tasks=num_tasks,
                num_nodes=num_nodes,
                node=node_id,
                global_rank=global_rank,
                local_rank=local_rank,
            )
        except (ImportError, RuntimeError, ValueError, KeyError):
            return None


class EnvironmentLSFInformationConfig(C.Config):
    """Configuration for LSF environment information."""

    hostname: str | None = None
    """The hostname of the current node."""

    hostnames: list[str] | None = None
    """List of hostnames for all nodes in the job."""

    job_id: str | None = None
    """The LSF job ID."""

    array_job_id: str | None = None
    """The LSF array job ID, if applicable."""

    array_task_id: str | None = None
    """The LSF array task ID, if applicable."""

    num_tasks: int | None = None
    """The number of tasks in the LSF job."""

    num_nodes: int | None = None
    """The number of nodes in the LSF job."""

    node: str | int | None = None
    """The node ID or name."""

    global_rank: int | None = None
    """The global rank of the current process."""

    local_rank: int | None = None
    """The local rank of the current process within its node."""

    @classmethod
    def empty(cls):
        return cls(
            hostname=None,
            hostnames=None,
            job_id=None,
            array_job_id=None,
            array_task_id=None,
            num_tasks=None,
            num_nodes=None,
            node=None,
            global_rank=None,
            local_rank=None,
        )

    @classmethod
    def from_current_environment(cls):
        try:
            import os
            import socket

            hostname = socket.gethostname()
            hostnames = [hostname]
            if node_list := os.environ.get("LSB_HOSTS", ""):
                hostnames = node_list.split()

            job_id = os.environ["LSB_JOBID"]
            array_job_id = os.environ.get("LSB_JOBINDEX")
            array_task_id = os.environ.get("LSB_JOBINDEX")

            num_tasks = int(os.environ.get("LSB_DJOB_NUMPROC", 1))
            num_nodes = len(set(hostnames))

            node_id = (
                os.environ.get("LSB_HOSTS", "").split().index(hostname)
                if "LSB_HOSTS" in os.environ
                else None
            )

            # LSF doesn't have direct equivalents for global_rank and local_rank
            # You might need to calculate these based on your specific setup
            global_rank = int(os.environ.get("PMI_RANK", 0))
            local_rank = int(os.environ.get("LSB_RANK", 0))

            return cls(
                hostname=hostname,
                hostnames=hostnames,
                job_id=job_id,
                array_job_id=array_job_id,
                array_task_id=array_task_id,
                num_tasks=num_tasks,
                num_nodes=num_nodes,
                node=node_id,
                global_rank=global_rank,
                local_rank=local_rank,
            )
        except (ImportError, RuntimeError, ValueError, KeyError):
            return None


class EnvironmentLinuxEnvironmentConfig(C.Config):
    """Configuration for Linux environment information."""

    user: str | None = None
    """The current user."""

    hostname: str | None = None
    """The hostname of the machine."""

    system: str | None = None
    """The operating system name."""

    release: str | None = None
    """The operating system release."""

    version: str | None = None
    """The operating system version."""

    machine: str | None = None
    """The machine type."""

    processor: str | None = None
    """The processor type."""

    cpu_count: int | None = None
    """The number of CPUs."""

    memory: int | None = None
    """The total system memory in bytes."""

    uptime: timedelta | None = None
    """The system uptime."""

    boot_time: float | None = None
    """The system boot time as a timestamp."""

    load_avg: tuple[float, float, float] | None = None
    """The system load average (1, 5, and 15 minutes)."""

    @classmethod
    def empty(cls):
        return cls(
            user=None,
            hostname=None,
            system=None,
            release=None,
            version=None,
            machine=None,
            processor=None,
            cpu_count=None,
            memory=None,
            uptime=None,
            boot_time=None,
            load_avg=None,
        )

    @classmethod
    def from_current_environment(cls):
        return cls(
            user=getpass.getuser(),
            hostname=platform.node(),
            system=platform.system(),
            release=platform.release(),
            version=platform.version(),
            machine=platform.machine(),
            processor=platform.processor(),
            cpu_count=os.cpu_count(),
            memory=psutil.virtual_memory().total,
            uptime=timedelta(seconds=psutil.boot_time()),
            boot_time=psutil.boot_time(),
            load_avg=os.getloadavg(),
        )


class EnvironmentSnapshotConfig(C.Config):
    """Configuration for environment snapshot information."""

    snapshot_dir: Path | None = None
    """The directory where the snapshot is stored."""

    modules: list[str] | None = None
    """List of modules included in the snapshot."""

    @classmethod
    def empty(cls):
        return cls(snapshot_dir=None, modules=None)

    @classmethod
    def from_current_environment(cls):
        draft = cls.draft()
        if snapshot_dir := os.environ.get("NSHRUNNER_SNAPSHOT_DIR"):
            draft.snapshot_dir = Path(snapshot_dir)
        if modules := os.environ.get("NSHRUNNER_SNAPSHOT_MODULES"):
            draft.modules = modules.split(",")
        return draft.finalize()


class EnvironmentPackageConfig(C.Config):
    """Configuration for Python package information."""

    name: str | None = None
    """The name of the package."""

    version: str | None = None
    """The version of the package."""

    path: Path | None = None
    """The installation path of the package."""

    summary: str | None = None
    """A brief summary of the package."""

    author: str | None = None
    """The author of the package."""

    license: str | None = None
    """The license of the package."""

    requires: list[str] | None = None
    """List of package dependencies."""

    @classmethod
    def empty(cls):
        return cls(
            name=None,
            version=None,
            path=None,
            summary=None,
            author=None,
            license=None,
            requires=None,
        )

    @classmethod
    def from_current_environment(cls):
        python_packages: dict[str, Self] = {}
        try:
            for dist in importlib.metadata.distributions():
                try:
                    # Get package metadata
                    metadata = dist.metadata

                    # Parse the version, stripping any local version identifier
                    pkg_version = version.parse(dist.version)
                    clean_version = (
                        f"{pkg_version.major}.{pkg_version.minor}.{pkg_version.micro}"
                    )

                    # Get requirements
                    requires = []
                    for req in dist.requires or []:
                        try:
                            requires.append(str(req))
                        except ValueError:
                            # If there's an invalid requirement, we'll skip it
                            log.warning(
                                f"Skipping invalid requirement for {dist.name}: {req}"
                            )

                    python_packages[dist.name] = cls(
                        name=dist.name,
                        version=clean_version,
                        path=Path(str(f)) if (f := dist.locate_file("")) else None,
                        summary=metadata["Summary"] if "Summary" in metadata else None,
                        author=metadata["Author"] if "Author" in metadata else None,
                        license=metadata["License"] if "License" in metadata else None,
                        requires=requires,
                    )
                except Exception:
                    log.warning(f"Error processing package {dist.name}", exc_info=True)

        except ImportError:
            log.warning(
                "importlib.metadata not available, skipping package information"
            )

        return python_packages


class EnvironmentGPUConfig(C.Config):
    """Configuration for individual GPU information."""

    name: str | None = None
    """Name of the GPU."""

    total_memory: int | None = None
    """Total memory of the GPU in bytes."""

    major: int | None = None
    """Major version of CUDA capability."""

    minor: int | None = None
    """Minor version of CUDA capability."""

    multi_processor_count: int | None = None
    """Number of multiprocessors on the GPU."""

    @classmethod
    def empty(cls):
        return cls(
            name=None,
            total_memory=None,
            major=None,
            minor=None,
            multi_processor_count=None,
        )


class EnvironmentCUDAConfig(C.Config):
    """Configuration for CUDA environment information."""

    is_available: bool | None = None
    """Whether CUDA is available."""

    version: str | None = None
    """CUDA version."""

    cudnn_version: int | None = None
    """cuDNN version."""

    @classmethod
    def empty(cls):
        return cls(is_available=None, version=None, cudnn_version=None)


class EnvironmentHardwareConfig(C.Config):
    """Configuration for hardware information."""

    cpu_count_physical: int | None = None
    """Number of physical CPU cores."""

    cpu_count_logical: int | None = None
    """Number of logical CPU cores."""

    cpu_frequency_current: float | None = None
    """Current CPU frequency in MHz."""

    cpu_frequency_min: float | None = None
    """Minimum CPU frequency in MHz."""

    cpu_frequency_max: float | None = None
    """Maximum CPU frequency in MHz."""

    ram_total: int | None = None
    """Total RAM in bytes."""

    ram_available: int | None = None
    """Available RAM in bytes."""

    disk_total: int | None = None
    """Total disk space in bytes."""

    disk_used: int | None = None
    """Used disk space in bytes."""

    disk_free: int | None = None
    """Free disk space in bytes."""

    gpu_count: int | None = None
    """Number of GPUs available."""

    gpus: list[EnvironmentGPUConfig] | None = None
    """List of GPU configurations."""

    cuda: EnvironmentCUDAConfig | None = None
    """CUDA environment configuration."""

    @classmethod
    def empty(cls):
        return cls(
            cpu_count_physical=None,
            cpu_count_logical=None,
            cpu_frequency_current=None,
            cpu_frequency_min=None,
            cpu_frequency_max=None,
            ram_total=None,
            ram_available=None,
            disk_total=None,
            disk_used=None,
            disk_free=None,
            gpu_count=None,
            gpus=None,
            cuda=None,
        )

    @classmethod
    def from_current_environment(cls):
        draft = cls.draft()

        # CPU information
        draft.cpu_count_physical = psutil.cpu_count(logical=False)
        draft.cpu_count_logical = psutil.cpu_count(logical=True)
        cpu_freq = psutil.cpu_freq()
        if cpu_freq:
            draft.cpu_frequency_current = cpu_freq.current
            draft.cpu_frequency_min = cpu_freq.min
            draft.cpu_frequency_max = cpu_freq.max

        # RAM information
        ram = psutil.virtual_memory()
        draft.ram_total = ram.total
        draft.ram_available = ram.available

        # Disk information
        disk = psutil.disk_usage("/")
        draft.disk_total = disk.total
        draft.disk_used = disk.used
        draft.disk_free = disk.free

        # GPU and CUDA information
        draft.cuda = EnvironmentCUDAConfig(
            is_available=torch.cuda.is_available(),
            version=cast(Any, torch).version.cuda,
            cudnn_version=torch.backends.cudnn.version()
            if torch.backends.cudnn.is_available()
            else None,
        )

        if draft.cuda.is_available:
            draft.gpu_count = torch.cuda.device_count()
            draft.gpus = []
            for i in range(draft.gpu_count):
                gpu_props = torch.cuda.get_device_properties(i)
                gpu_config = EnvironmentGPUConfig(
                    name=gpu_props.name,
                    total_memory=gpu_props.total_memory,
                    major=gpu_props.major,
                    minor=gpu_props.minor,
                    multi_processor_count=gpu_props.multi_processor_count,
                )
                draft.gpus.append(gpu_config)

        return draft.finalize()


class GitRepositoryConfig(C.Config):
    """Configuration for Git repository information."""

    is_git_repo: bool | None = None
    """Whether the current directory is a Git repository."""

    branch: str | None = None
    """The current Git branch."""

    commit_hash: str | None = None
    """The current commit hash."""

    commit_message: str | None = None
    """The current commit message."""

    author: str | None = None
    """The author of the current commit."""

    commit_date: str | None = None
    """The date of the current commit."""

    remote_url: str | None = None
    """The URL of the remote repository."""

    is_dirty: bool | None = None
    """Whether there are uncommitted changes."""

    @classmethod
    def empty(cls):
        return cls(
            is_git_repo=None,
            branch=None,
            commit_hash=None,
            commit_message=None,
            author=None,
            commit_date=None,
            remote_url=None,
            is_dirty=None,
        )

    @classmethod
    def from_current_directory(cls):
        try:
            import git
        except ImportError:
            return cls()

        draft = cls.draft()
        try:
            repo = git.Repo(os.getcwd(), search_parent_directories=True)
            draft.is_git_repo = True
            draft.branch = repo.active_branch.name
            commit = repo.head.commit
            draft.commit_hash = commit.hexsha

            # Handle both str and bytes for commit message
            if isinstance(commit.message, str):
                draft.commit_message = commit.message.strip()
            elif isinstance(commit.message, bytes):
                draft.commit_message = commit.message.decode(
                    "utf-8", errors="replace"
                ).strip()
            else:
                draft.commit_message = str(commit.message).strip()

            draft.author = f"{commit.author.name} <{commit.author.email}>"
            draft.commit_date = commit.committed_datetime.isoformat()
            if repo.remotes:
                draft.remote_url = repo.remotes.origin.url
            draft.is_dirty = repo.is_dirty()
        except git.InvalidGitRepositoryError:
            draft.is_git_repo = False
        except Exception:
            log.warning("Failed to get Git repository information", exc_info=True)
            draft.is_git_repo = None

        return draft.finalize()


class EnvironmentConfig(C.Config):
    """Configuration for the overall environment."""

    cwd: Path | None = None
    """The current working directory."""

    snapshot: EnvironmentSnapshotConfig | None = None
    """The environment snapshot configuration."""

    python_executable: Path | None = None
    """The path to the Python executable."""

    python_path: list[Path] | None = None
    """The Python path."""

    python_version: str | None = None
    """The Python version."""

    python_packages: dict[str, EnvironmentPackageConfig] | None = None
    """A mapping of package names to their configurations."""

    config: EnvironmentClassInformationConfig | None = None
    """The configuration class information."""

    model: EnvironmentClassInformationConfig | None = None
    """The Lightning module class information."""

    datamodule: EnvironmentClassInformationConfig | None = None
    """The Lightning data module class information."""

    linux: EnvironmentLinuxEnvironmentConfig | None = None
    """The Linux environment information."""

    hardware: EnvironmentHardwareConfig | None = None
    """Hardware configuration information."""

    slurm: EnvironmentSLURMInformationConfig | None = None
    """The SLURM environment information."""

    lsf: EnvironmentLSFInformationConfig | None = None
    """The LSF environment information."""

    base_dir: Path | None = None
    """The base directory for the run."""

    log_dir: Path | None = None
    """The directory for logs."""

    checkpoint_dir: Path | None = None
    """The directory for checkpoints."""

    stdio_dir: Path | None = None
    """The directory for standard input/output files."""

    seed: int | None = None
    """The global random seed."""

    seed_workers: bool | None = None
    """Whether to seed workers."""

    git: GitRepositoryConfig | None = None
    """Git repository information."""

    @classmethod
    def empty(cls):
        return cls(
            cwd=None,
            snapshot=None,
            python_executable=None,
            python_path=None,
            python_version=None,
            python_packages=None,
            config=None,
            model=None,
            linux=None,
            hardware=None,
            slurm=None,
            lsf=None,
            base_dir=None,
            log_dir=None,
            checkpoint_dir=None,
            stdio_dir=None,
            seed=None,
            seed_workers=None,
            git=None,
        )

    @classmethod
    def from_current_environment(
        cls,
        trainer_config: TrainerConfig,
        model: LightningModule,
        datamodule: LightningDataModule | None = None,
    ):
        draft = cls.draft()
        draft.cwd = Path(os.getcwd())
        draft.python_executable = Path(sys.executable)
        draft.python_path = [Path(path) for path in sys.path]
        draft.python_version = sys.version
        draft.python_packages = EnvironmentPackageConfig.from_current_environment()
        draft.config = EnvironmentClassInformationConfig.from_instance(trainer_config)
        draft.model = EnvironmentClassInformationConfig.from_instance(model)
        if datamodule is not None:
            draft.datamodule = EnvironmentClassInformationConfig.from_instance(
                datamodule
            )
        draft.linux = EnvironmentLinuxEnvironmentConfig.from_current_environment()
        draft.hardware = EnvironmentHardwareConfig.from_current_environment()
        draft.slurm = EnvironmentSLURMInformationConfig.from_current_environment()
        draft.lsf = EnvironmentLSFInformationConfig.from_current_environment()
        draft.base_dir = trainer_config.directory.resolve_run_root_directory(
            trainer_config.id
        )
        draft.log_dir = trainer_config.directory.resolve_subdirectory(
            trainer_config.id, "log"
        )
        draft.checkpoint_dir = trainer_config.directory.resolve_subdirectory(
            trainer_config.id, "checkpoint"
        )
        draft.stdio_dir = trainer_config.directory.resolve_subdirectory(
            trainer_config.id, "stdio"
        )
        draft.seed = (
            int(seed_str) if (seed_str := os.environ.get("PL_GLOBAL_SEED")) else None
        )
        draft.seed_workers = (
            bool(int(seed_everything))
            if (seed_everything := os.environ.get("PL_SEED_WORKERS"))
            else None
        )
        draft.snapshot = EnvironmentSnapshotConfig.from_current_environment()
        draft.git = GitRepositoryConfig.from_current_directory()
        return draft.finalize()



from __future__ import annotations

__codegen__ = True

from nshtrainer.callbacks.early_stopping import CallbackConfigBase as CallbackConfigBase
from nshtrainer.callbacks.early_stopping import (
    EarlyStoppingCallbackConfig as EarlyStoppingCallbackConfig,
)
from nshtrainer.callbacks.early_stopping import MetricConfig as MetricConfig
from nshtrainer.callbacks.early_stopping import callback_registry as callback_registry

__all__ = [
    "CallbackConfigBase",
    "EarlyStoppingCallbackConfig",
    "MetricConfig",
    "callback_registry",
]



from __future__ import annotations

from .best_checkpoint import BestCheckpointCallback as BestCheckpointCallback
from .best_checkpoint import (
    BestCheckpointCallbackConfig as BestCheckpointCallbackConfig,
)
from .last_checkpoint import LastCheckpointCallback as LastCheckpointCallback
from .last_checkpoint import (
    LastCheckpointCallbackConfig as LastCheckpointCallbackConfig,
)
from .on_exception_checkpoint import (
    OnExceptionCheckpointCallback as OnExceptionCheckpointCallback,
)
from .on_exception_checkpoint import (
    OnExceptionCheckpointCallbackConfig as OnExceptionCheckpointCallbackConfig,
)



from __future__ import annotations

import errno
import hashlib
import logging
import os
import platform
import shutil
from pathlib import Path

from typing_extensions import TypeAliasType

log = logging.getLogger(__name__)

_Path = TypeAliasType("_Path", str | Path)


def get_relative_path(source: _Path, destination: _Path):
    # Get the absolute paths
    source = os.path.abspath(source)
    destination = os.path.abspath(destination)

    # Split the paths into components
    source_parts = source.split(os.sep)
    destination_parts = destination.split(os.sep)

    # Find the point where the paths diverge
    i = 0
    for i in range(min(len(source_parts), len(destination_parts))):
        if source_parts[i] != destination_parts[i]:
            break
    else:
        i += 1

    # Build the relative path
    up = os.sep.join([".." for _ in range(len(source_parts) - i - 1)])
    down = os.sep.join(destination_parts[i:])

    return Path(os.path.normpath(os.path.join(up, down)))


def find_symlinks(
    target_file: _Path,
    *search_directories: _Path,
    glob_pattern: str = "*",
):
    target_file = Path(target_file).resolve()
    symlinks: list[Path] = []

    for search_directory in search_directories:
        search_directory = Path(search_directory)
        for path in search_directory.rglob(glob_pattern):
            if path.is_symlink():
                try:
                    link_target = path.resolve()
                    if link_target.samefile(target_file):
                        symlinks.append(path)
                except FileNotFoundError:
                    # Handle broken symlinks
                    pass

    return symlinks


def compute_file_checksum(file_path: Path) -> str:
    """
    Calculate the SHA256 checksum of a file.

    Args:
        file_path (Path): The path to the file.

    Returns:
        str: The hexadecimal representation of the file's SHA256 checksum.
    """
    sha256_hash = hashlib.sha256()
    with file_path.open("rb") as f:
        for byte_block in iter(lambda: f.read(4096), b""):
            sha256_hash.update(byte_block)
    return sha256_hash.hexdigest()


def try_symlink_or_copy(
    file_path: Path,
    link_path: Path,
    *,
    target_is_directory: bool = False,
    relative: bool = True,
    remove_existing: bool = True,
    throw_on_invalid_target: bool = False,
):
    """
    Symlinks on Unix, copies on Windows.
    """

    # Check if the target file exists
    if throw_on_invalid_target and not file_path.exists():
        raise FileNotFoundError(f"File not found: {file_path}")

    # If the link already exists, remove it
    if remove_existing:
        try:
            if path_exists(link_path, follow_symlinks=False):
                # follow_symlinks=False is EXTREMELY important here
                # Otherwise, we've already deleted the file that the symlink
                # used to point to, so this always returns False
                if link_path.is_dir():
                    shutil.rmtree(link_path)
                else:
                    link_path.unlink(missing_ok=True)
        except Exception:
            log.warning(f"Failed to remove {link_path}", exc_info=True)
        else:
            log.debug(f"Removed {link_path=}")

    symlink_target = get_relative_path(link_path, file_path) if relative else file_path
    try:
        if platform.system() == "Windows":
            if target_is_directory:
                shutil.copytree(file_path, link_path)
            else:
                shutil.copy(file_path, link_path)
        else:
            link_path.symlink_to(
                symlink_target,
                target_is_directory=target_is_directory,
            )
    except Exception:
        log.warning(
            f"Failed to create symlink or copy {file_path} to {link_path}",
            exc_info=True,
        )
        return False
    else:
        log.debug(f"Created symlink or copied {file_path} to {link_path}")
        return True


_WINERROR_NOT_READY = 21  # drive exists but is not accessible
_WINERROR_INVALID_NAME = 123  # fix for bpo-35306
_WINERROR_CANT_RESOLVE_FILENAME = 1921  # broken symlink pointing to itself

# EBADF - guard against macOS `stat` throwing EBADF
_IGNORED_ERRNOS = (errno.ENOENT, errno.ENOTDIR, errno.EBADF, errno.ELOOP)

_IGNORED_WINERRORS = (
    _WINERROR_NOT_READY,
    _WINERROR_INVALID_NAME,
    _WINERROR_CANT_RESOLVE_FILENAME,
)


def _ignore_error(exception):
    return (
        getattr(exception, "errno", None) in _IGNORED_ERRNOS
        or getattr(exception, "winerror", None) in _IGNORED_WINERRORS
    )


def path_exists(path: Path, follow_symlinks: bool = True):
    """
    Whether this path exists.

    This method normally follows symlinks; to check whether a symlink exists,
    add the argument follow_symlinks=False.
    """
    try:
        path.stat(follow_symlinks=follow_symlinks)
    except OSError as e:
        if not _ignore_error(e):
            raise
        return False
    except ValueError:
        # Non-encodable path
        return False
    return True



from __future__ import annotations

__codegen__ = True

from nshtrainer.profiler.pytorch import BaseProfilerConfig as BaseProfilerConfig
from nshtrainer.profiler.pytorch import PyTorchProfilerConfig as PyTorchProfilerConfig

__all__ = [
    "BaseProfilerConfig",
    "PyTorchProfilerConfig",
]



from __future__ import annotations

__codegen__ = True

from nshtrainer.util._environment_info import (
    EnvironmentClassInformationConfig as EnvironmentClassInformationConfig,
)
from nshtrainer.util._environment_info import EnvironmentConfig as EnvironmentConfig
from nshtrainer.util._environment_info import (
    EnvironmentCUDAConfig as EnvironmentCUDAConfig,
)
from nshtrainer.util._environment_info import (
    EnvironmentGPUConfig as EnvironmentGPUConfig,
)
from nshtrainer.util._environment_info import (
    EnvironmentHardwareConfig as EnvironmentHardwareConfig,
)
from nshtrainer.util._environment_info import (
    EnvironmentLinuxEnvironmentConfig as EnvironmentLinuxEnvironmentConfig,
)
from nshtrainer.util._environment_info import (
    EnvironmentLSFInformationConfig as EnvironmentLSFInformationConfig,
)
from nshtrainer.util._environment_info import (
    EnvironmentPackageConfig as EnvironmentPackageConfig,
)
from nshtrainer.util._environment_info import (
    EnvironmentSLURMInformationConfig as EnvironmentSLURMInformationConfig,
)
from nshtrainer.util._environment_info import (
    EnvironmentSnapshotConfig as EnvironmentSnapshotConfig,
)
from nshtrainer.util._environment_info import GitRepositoryConfig as GitRepositoryConfig

__all__ = [
    "EnvironmentCUDAConfig",
    "EnvironmentClassInformationConfig",
    "EnvironmentConfig",
    "EnvironmentGPUConfig",
    "EnvironmentHardwareConfig",
    "EnvironmentLSFInformationConfig",
    "EnvironmentLinuxEnvironmentConfig",
    "EnvironmentPackageConfig",
    "EnvironmentSLURMInformationConfig",
    "EnvironmentSnapshotConfig",
    "GitRepositoryConfig",
]



from __future__ import annotations

__codegen__ = True

from nshtrainer.lr_scheduler.base import LRSchedulerConfigBase as LRSchedulerConfigBase
from nshtrainer.lr_scheduler.base import lr_scheduler_registry as lr_scheduler_registry

__all__ = [
    "LRSchedulerConfigBase",
    "lr_scheduler_registry",
]



from __future__ import annotations

__codegen__ = True

from nshtrainer.loggers.csv import CSVLoggerConfig as CSVLoggerConfig
from nshtrainer.loggers.csv import LoggerConfigBase as LoggerConfigBase
from nshtrainer.loggers.csv import logger_registry as logger_registry

__all__ = [
    "CSVLoggerConfig",
    "LoggerConfigBase",
    "logger_registry",
]



from __future__ import annotations

__codegen__ = True

from nshtrainer.loggers import ActSaveLoggerConfig as ActSaveLoggerConfig
from nshtrainer.loggers import CSVLoggerConfig as CSVLoggerConfig
from nshtrainer.loggers import LoggerConfig as LoggerConfig
from nshtrainer.loggers import LoggerConfigBase as LoggerConfigBase
from nshtrainer.loggers import TensorboardLoggerConfig as TensorboardLoggerConfig
from nshtrainer.loggers import WandbLoggerConfig as WandbLoggerConfig
from nshtrainer.loggers import logger_registry as logger_registry
from nshtrainer.loggers.wandb import CallbackConfigBase as CallbackConfigBase
from nshtrainer.loggers.wandb import (
    WandbUploadCodeCallbackConfig as WandbUploadCodeCallbackConfig,
)
from nshtrainer.loggers.wandb import (
    WandbWatchCallbackConfig as WandbWatchCallbackConfig,
)

from . import actsave as actsave
from . import base as base
from . import csv as csv
from . import tensorboard as tensorboard
from . import wandb as wandb

__all__ = [
    "ActSaveLoggerConfig",
    "CSVLoggerConfig",
    "CallbackConfigBase",
    "LoggerConfig",
    "LoggerConfigBase",
    "TensorboardLoggerConfig",
    "WandbLoggerConfig",
    "WandbUploadCodeCallbackConfig",
    "WandbWatchCallbackConfig",
    "actsave",
    "base",
    "csv",
    "logger_registry",
    "tensorboard",
    "wandb",
]



from __future__ import annotations

import signal
from typing import Literal

from typing_extensions import TypeAliasType, override

from .base import PluginConfigBase, plugin_registry


@plugin_registry.register
class KubeflowEnvironmentPlugin(PluginConfigBase):
    """Environment for distributed training using the PyTorchJob operator from Kubeflow.

    This environment, unlike others, does not get auto-detected and needs to be passed
    to the Fabric/Trainer constructor manually.
    """

    name: Literal["kubeflow_environment"] = "kubeflow_environment"

    @override
    def create_plugin(self, trainer_config):
        from lightning.fabric.plugins.environments.kubeflow import KubeflowEnvironment

        return KubeflowEnvironment()


@plugin_registry.register
class LightningEnvironmentPlugin(PluginConfigBase):
    """The default environment used by Lightning for a single node or free cluster (not managed).

    There are two modes the Lightning environment can operate with:
    1. User launches main process by `python train.py ...` with no additional environment variables.
       Lightning will spawn new worker processes for distributed training in the current node.
    2. User launches all processes manually or with utilities like `torch.distributed.launch`.
       The appropriate environment variables need to be set, and at minimum `LOCAL_RANK`.
    """

    name: Literal["lightning_environment"] = "lightning_environment"

    @override
    def create_plugin(self, trainer_config):
        from lightning.fabric.plugins.environments.lightning import LightningEnvironment

        return LightningEnvironment()


@plugin_registry.register
class LSFEnvironmentPlugin(PluginConfigBase):
    """An environment for running on clusters managed by the LSF resource manager.

    It is expected that any execution using this ClusterEnvironment was executed
    using the Job Step Manager i.e. `jsrun`.
    """

    name: Literal["lsf_environment"] = "lsf_environment"

    @override
    def create_plugin(self, trainer_config):
        from lightning.fabric.plugins.environments.lsf import LSFEnvironment

        return LSFEnvironment()


@plugin_registry.register
class MPIEnvironmentPlugin(PluginConfigBase):
    """An environment for running on clusters with processes created through MPI.

    Requires the installation of the `mpi4py` package.
    """

    name: Literal["mpi_environment"] = "mpi_environment"

    @override
    def create_plugin(self, trainer_config):
        from lightning.fabric.plugins.environments.mpi import MPIEnvironment

        return MPIEnvironment()


SignalAlias = TypeAliasType(
    "SignalAlias",
    Literal[
        "SIGABRT",
        "SIGFPE",
        "SIGILL",
        "SIGINT",
        "SIGSEGV",
        "SIGTERM",
        "SIGBREAK",
        "CTRL_C_EVENT",
        "CTRL_BREAK_EVENT",
        "SIGALRM",
        "SIGBUS",
        "SIGCHLD",
        "SIGCONT",
        "SIGHUP",
        "SIGIO",
        "SIGIOT",
        "SIGKILL",
        "SIGPIPE",
        "SIGPROF",
        "SIGQUIT",
        "SIGSTOP",
        "SIGSYS",
        "SIGTRAP",
        "SIGTSTP",
        "SIGTTIN",
        "SIGTTOU",
        "SIGURG",
        "SIGUSR1",
        "SIGUSR2",
        "SIGVTALRM",
        "SIGWINCH",
        "SIGXCPU",
        "SIGXFSZ",
        "SIGEMT",
        "SIGINFO",
        "SIGCLD",
        "SIGPOLL",
        "SIGPWR",
        "SIGRTMAX",
        "SIGRTMIN",
        "SIGSTKFLT",
    ],
)


@plugin_registry.register
class SLURMEnvironmentPlugin(PluginConfigBase):
    """An environment for running on clusters managed by the SLURM resource manager."""

    name: Literal["slurm_environment"] = "slurm_environment"

    auto_requeue: bool = True
    """Whether automatic job resubmission is enabled or not."""

    requeue_signal: SignalAlias | None = None
    """The signal that SLURM will send to indicate that the job should be requeued."""

    @override
    def create_plugin(self, trainer_config):
        from lightning.fabric.plugins.environments.slurm import SLURMEnvironment

        requeue_signal = None
        if self.requeue_signal is not None:
            try:
                requeue_signal = signal.Signals[self.requeue_signal]
            except KeyError:
                raise ValueError(
                    f"Invalid signal name: {self.requeue_signal}. "
                    "Please provide a valid signal name from the signal module."
                )

        return SLURMEnvironment(
            auto_requeue=self.auto_requeue,
            requeue_signal=requeue_signal,
        )


@plugin_registry.register
class TorchElasticEnvironmentPlugin(PluginConfigBase):
    """Environment for fault-tolerant and elastic training with torchelastic."""

    name: Literal["torchelastic_environment"] = "torchelastic_environment"

    @override
    def create_plugin(self, trainer_config):
        from lightning.fabric.plugins.environments.torchelastic import (
            TorchElasticEnvironment,
        )

        return TorchElasticEnvironment()


@plugin_registry.register
class XLAEnvironmentPlugin(PluginConfigBase):
    """Cluster environment for training on a TPU Pod with the PyTorch/XLA library."""

    name: Literal["xla_environment"] = "xla_environment"

    @override
    def create_plugin(self, trainer_config):
        from lightning.fabric.plugins.environments.xla import XLAEnvironment

        return XLAEnvironment()



from __future__ import annotations

__codegen__ = True

from nshtrainer._hf_hub import CallbackConfigBase as CallbackConfigBase
from nshtrainer._hf_hub import (
    HuggingFaceHubAutoCreateConfig as HuggingFaceHubAutoCreateConfig,
)
from nshtrainer._hf_hub import HuggingFaceHubConfig as HuggingFaceHubConfig
from nshtrainer._hf_hub import callback_registry as callback_registry

__all__ = [
    "CallbackConfigBase",
    "HuggingFaceHubAutoCreateConfig",
    "HuggingFaceHubConfig",
    "callback_registry",
]



from __future__ import annotations

__codegen__ = True

from nshtrainer.callbacks.checkpoint._base import (
    BaseCheckpointCallbackConfig as BaseCheckpointCallbackConfig,
)
from nshtrainer.callbacks.checkpoint._base import (
    CallbackConfigBase as CallbackConfigBase,
)
from nshtrainer.callbacks.checkpoint._base import (
    CheckpointMetadata as CheckpointMetadata,
)

__all__ = [
    "BaseCheckpointCallbackConfig",
    "CallbackConfigBase",
    "CheckpointMetadata",
]



from __future__ import annotations

import logging
from typing import Literal

import nshconfig as C
from typing_extensions import final, override

from .base import LoggerConfigBase, logger_registry

log = logging.getLogger(__name__)


def _tensorboard_available():
    try:
        from lightning.fabric.loggers.tensorboard import (
            _TENSORBOARD_AVAILABLE,
            _TENSORBOARDX_AVAILABLE,
        )

        if not _TENSORBOARD_AVAILABLE and not _TENSORBOARDX_AVAILABLE:
            log.warning(
                "TensorBoard/TensorBoardX not found. Disabling TensorBoardLogger. "
                "Please install TensorBoard with `pip install tensorboard` or "
                "TensorBoardX with `pip install tensorboardx` to enable TensorBoard logging."
            )
            return False
        return True
    except ImportError:
        return False


@final
@logger_registry.register
class TensorboardLoggerConfig(LoggerConfigBase):
    name: Literal["tensorboard"] = "tensorboard"

    enabled: bool = C.Field(default_factory=lambda: _tensorboard_available())
    """Enable TensorBoard logging."""

    priority: int = 2
    """Priority of the logger. Higher priority loggers are created first."""

    log_graph: bool = False
    """
    Adds the computational graph to tensorboard. This requires that
        the user has defined the `self.example_input_array` attribute in their
        model.
    """

    default_hp_metric: bool = True
    """
    Enables a placeholder metric with key `hp_metric` when `log_hyperparams` is
        called without a metric (otherwise calls to log_hyperparams without a metric are ignored).
    """

    prefix: str = ""
    """A string to put at the beginning of metric keys."""

    @override
    def create_logger(self, trainer_config):
        if not self.enabled:
            return None

        from lightning.pytorch.loggers.tensorboard import TensorBoardLogger

        save_dir = trainer_config.directory._resolve_log_directory_for_logger(
            trainer_config.id,
            self,
        )
        return TensorBoardLogger(
            save_dir=save_dir,
            name=trainer_config.full_name,
            version=trainer_config.id,
            log_graph=self.log_graph,
            default_hp_metric=self.default_hp_metric,
        )



from __future__ import annotations

import contextlib
import logging
import os
import re
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar, Literal, cast

import nshconfig as C
from nshrunner._env import SNAPSHOT_DIR
from typing_extensions import assert_never, override

from ._callback import NTCallbackBase
from .callbacks.base import (
    CallbackConfigBase,
    CallbackMetadataConfig,
    callback_registry,
)

if TYPE_CHECKING:
    from huggingface_hub import HfApi  # noqa: F401

    from .trainer._config import TrainerConfig


log = logging.getLogger(__name__)


class HuggingFaceHubAutoCreateConfig(C.Config):
    enabled: bool = True
    """Enable automatic repository creation on the Hugging Face Hub."""

    private: bool = True
    """Whether to create the repository as private."""

    namespace: str | None = None
    """The namespace to create the repository in. If `None`, the repository will be created in the user's namespace."""

    def __bool__(self):
        return self.enabled


@callback_registry.register
class HuggingFaceHubConfig(CallbackConfigBase):
    """Configuration options for Hugging Face Hub integration."""

    name: Literal["hf_hub"] = "hf_hub"

    metadata: ClassVar[CallbackMetadataConfig] = {"ignore_if_exists": True}

    enabled: bool = False
    """Enable Hugging Face Hub integration."""

    token: str | None = None
    """Hugging Face Hub API token. If `None`, the token will be read from the current environment.
    This needs to either be set using `huggingface-cli login` or by setting the `HUGGINGFACE_TOKEN`
    environment variable."""

    auto_create: HuggingFaceHubAutoCreateConfig = HuggingFaceHubAutoCreateConfig()
    """Automatic repository creation configuration options."""

    save_config: bool = True
    """Whether to save the model configuration to the Hugging Face Hub."""

    save_checkpoints: bool = True
    """Whether to save checkpoints to the Hugging Face Hub."""

    save_code: bool = True
    """Whether to save code to the Hugging Face Hub.
    This is only supported if `nshsnap` is installed and snapshotting is enabled."""

    save_in_background: bool = True
    """Whether to save to the Hugging Face Hub in the background.
    This corresponds to setting `run_as_future=True` in the HFApi upload methods."""

    on_login_error: Literal["warn", "error", "ignore"] = "warn"
    """What to do when an error occurs during login."""

    def enable_(self):
        self.enabled = True
        return self

    def disable_(self):
        self.enabled = False
        return self

    def __bool__(self):
        return self.enabled

    @override
    def create_callbacks(self, trainer_config):
        # Attempt to login. If it fails, we'll log a warning or error based on the configuration.
        try:
            api = _api(self.token)
            if api is None:
                raise ValueError("Failed to create Hugging Face Hub API instance.")
        except Exception as e:
            match self.on_login_error:
                case "warn":
                    log.warning(
                        "Failed to create Hugging Face Hub API instance. Disabling Hugging Face Hub integration.",
                        exc_info=e,
                    )
                    return
                case "error":
                    raise
                case "ignore":
                    log.debug(
                        "Failed to create Hugging Face Hub API instance. Disabling Hugging Face Hub integration.",
                        exc_info=e,
                    )
                    return
                case _:
                    assert_never(self.on_login_error)

        yield HFHubCallback(self)


def _api(token: str | None = None):
    # Make sure that `huggingface_hub` is installed
    try:
        import huggingface_hub  # noqa: F401
    except ImportError:
        log.exception(
            "Could not import `huggingface_hub`. Please install it using `pip install huggingface_hub`."
        )
        return None

    # Create and authenticate the API instance
    try:
        api = huggingface_hub.HfApi(token=token)

        # Verify authentication
        api.whoami()
    except Exception:
        log.exception(
            "Authentication failed for Hugging Face Hub. "
            "Please make sure you are logged in using `huggingface-cli login`, "
            "by setting the HUGGING_FACE_HUB_TOKEN environment variable, "
            "or by providing a valid token in the configuration."
        )
        return None

    return api


def _repo_name(api: HfApi, trainer_config: TrainerConfig):
    username = None
    if (ac := trainer_config.hf_hub.auto_create) and ac.namespace:
        username = ac.namespace
    elif (username := api.whoami().get("name", None)) is None:
        raise ValueError("Could not get username from Hugging Face Hub.")

    # Sanitize the project (if it exists), run_name, and id
    parts = []
    if trainer_config.project:
        parts.append(re.sub(r"[^a-zA-Z0-9-]", "-", trainer_config.project))
    if trainer_config.full_name:
        parts.append(re.sub(r"[^a-zA-Z0-9-]", "-", trainer_config.full_name))
    parts.append(re.sub(r"[^a-zA-Z0-9-]", "-", trainer_config.id))

    # Combine parts and ensure it starts and ends with alphanumeric characters
    repo_name = "-".join(parts)
    repo_name = repo_name.strip("-")
    repo_name = re.sub(
        r"-+", "-", repo_name
    )  # Replace multiple dashes with a single dash

    # Ensure the name is not longer than 96 characters (excluding username)
    if len(repo_name) > 96:
        repo_name = repo_name[:96].rstrip("-")

    # Ensure the repo name starts with an alphanumeric character
    repo_name = re.sub(r"^[^a-zA-Z0-9]+", "", repo_name)

    # If the repo_name is empty after all sanitization, use a default name
    if not repo_name:
        repo_name = "default-repo-name"

    return f"{username}/{repo_name}"


@dataclass
class _Upload:
    local_path: Path
    path_in_repo: Path

    @classmethod
    def from_local_path(cls, local_path: Path, trainer_config: TrainerConfig):
        # Resolve the checkpoint directory
        checkpoint_dir = trainer_config.directory.resolve_subdirectory(
            trainer_config.id, "checkpoint"
        )

        try:
            relative_path = local_path.relative_to(checkpoint_dir)
        except ValueError:
            raise ValueError(
                f"Checkpoint path {local_path} is not within the checkpoint directory {checkpoint_dir}."
            )

        # Prefix the path in repo with "checkpoints"
        path_in_repo = Path("checkpoints") / relative_path

        return cls(local_path=local_path, path_in_repo=path_in_repo)


class HFHubCallback(NTCallbackBase):
    @contextlib.contextmanager
    def _with_error_handling(self, opeartion: str):
        try:
            yield
        except Exception:
            log.warning(
                f"Failed to {opeartion}, repo_id={self._repo_id}", exc_info=True
            )
        else:
            log.debug(f"Successfully {opeartion}, repo_id={self._repo_id}")

    def __init__(self, config: HuggingFaceHubConfig):
        super().__init__()

        self.config = config

        self._repo_id: str | None = None
        self._checksum_to_path_in_repo: dict[str, Path] = {}

    @override
    def setup(self, trainer, pl_module, stage):
        self._repo_id = _repo_name(self.api, trainer.hparams)

        if not self.config or not trainer.is_global_zero:
            return

        # Create the repository, if it doesn't exist
        self._create_repo_if_not_exists()

        # Upload the config and code
        self._save_config(trainer.hparams)
        self._save_code()

    @override
    def on_checkpoint_saved(self, ckpt_path, metadata_path, trainer, pl_module):
        # If HF Hub is enabled, then we upload
        if (
            not self.config
            or not self.config.save_checkpoints
            or not trainer.is_global_zero
        ):
            return

        with self._with_error_handling("save checkpoints"):
            self._save_checkpoint(
                _Upload.from_local_path(ckpt_path, trainer.hparams),
                _Upload.from_local_path(metadata_path, trainer.hparams)
                if metadata_path is not None
                else None,
            )

    @cached_property
    def api(self):
        # Create and authenticate the API instance
        if (api := _api(self.config.token)) is None:
            raise ValueError("Failed to create Hugging Face Hub API instance.")
        return api

    @property
    def repo_id(self):
        if self._repo_id is None:
            raise ValueError("Repository id has not been initialized.")
        return self._repo_id

    def _create_repo_if_not_exists(self):
        if not self.config or not self.config.auto_create:
            return

        # Create the repository, if it doesn't exist
        with self._with_error_handling("create repository"):
            from huggingface_hub.utils import RepositoryNotFoundError

            try:
                # Check if the repository exists
                self.api.repo_info(repo_id=self.repo_id, repo_type="model")
                log.info(f"Repository '{self.repo_id}' already exists.")
            except RepositoryNotFoundError:
                # Repository doesn't exist, so create it
                try:
                    self.api.create_repo(
                        repo_id=self.repo_id,
                        repo_type="model",
                        private=self.config.auto_create.private,
                        exist_ok=True,
                    )
                    log.info(f"Created new repository '{self.repo_id}'.")
                except Exception:
                    log.warning(
                        f"Failed to create repository '{self.repo_id}'", exc_info=True
                    )
            except Exception:
                log.warning(
                    f"Error checking repository '{self.repo_id}'", exc_info=True
                )

    def _save_config(self, trainer_config: TrainerConfig):
        with self._with_error_handling("upload config"):
            self.api.upload_file(
                path_or_fileobj=trainer_config.model_dump_json(indent=4).encode(
                    "utf-8"
                ),
                path_in_repo="config.json",
                repo_id=self.repo_id,
                repo_type="model",
                run_as_future=cast(Any, self.config.save_in_background),
            )

    def _save_code(self):
        # If a snapshot has been taken (which can be detected using the SNAPSHOT_DIR env),
        # then upload all contents within the snapshot directory to the repository.
        if not (snapshot_dir := os.environ.get(SNAPSHOT_DIR)):
            log.debug("No snapshot directory found. Skipping upload.")
            return

        with self._with_error_handling("save code"):
            snapshot_dir = Path(snapshot_dir)
            if not snapshot_dir.exists() or not snapshot_dir.is_dir():
                log.warning(
                    f"Snapshot directory '{snapshot_dir}' does not exist or is not a directory."
                )
                return

            self.api.upload_folder(
                folder_path=str(snapshot_dir),
                repo_id=self.repo_id,
                repo_type="model",
                path_in_repo="code",  # Prefix with "code" folder
                run_as_future=cast(Any, self.config.save_in_background),
            )

    def _save_file(self, p: _Upload):
        with self._with_error_handling("save file"):
            # First, read the file into memory.
            # We do this to avoid issues with
            # the file being moved or deleted.
            with p.local_path.open("rb") as f:
                data = f.read()

            # Upload the checkpoint files to the repository
            self.api.upload_file(
                path_or_fileobj=data,
                path_in_repo=str(p.path_in_repo),
                repo_id=self.repo_id,
                repo_type="model",
                run_as_future=cast(Any, self.config.save_in_background),
            )

    def _copy_file(self, source_path_in_repo: Path, dest_path_in_repo: Path):
        # Create a commit for copying the files
        from huggingface_hub.hf_api import CommitOperationCopy

        with self._with_error_handling("copy file"):
            copy_op = CommitOperationCopy(
                src_path_in_repo=str(source_path_in_repo),
                path_in_repo=str(dest_path_in_repo),
            )

            self.api.create_commit(
                repo_id=self.repo_id,
                repo_type="model",
                commit_message="Copy checkpoint file",
                operations=[copy_op],
                run_as_future=cast(Any, self.config.save_in_background),
            )

    def _save_checkpoint(self, path: _Upload, metadata_path: _Upload | None):
        if not self.config.save_checkpoints:
            return

        # If no metadata, just save regularly.
        if metadata_path is None:
            self._save_file(path)
            return

        # Otherwise, let's check to see if we've already uploaded the metadata.
        # If so, we can just copy the checkpoint file.
        from ._checkpoint.metadata import CheckpointMetadata

        metadata = CheckpointMetadata.from_file(metadata_path.local_path)
        if (
            existing_ckpt_path := self._checksum_to_path_in_repo.get(
                metadata.checkpoint_checksum
            )
        ) is not None:
            self._copy_file(existing_ckpt_path, path.path_in_repo)
        else:
            # Otherwise, we save the checkpoint & keep the checksum so we don't
            # re-upload the same file again.
            self._save_file(path)
            self._checksum_to_path_in_repo[metadata.checkpoint_checksum] = (
                path.path_in_repo
            )

        # Save the metadata file
        # NOTE: This file is fairly small, so we can just upload it directly.
        # No need to copy.
        self._save_file(metadata_path)



from __future__ import annotations

__codegen__ = True

from nshtrainer.trainer.accelerator import AcceleratorConfig as AcceleratorConfig
from nshtrainer.trainer.accelerator import (
    AcceleratorConfigBase as AcceleratorConfigBase,
)
from nshtrainer.trainer.accelerator import CPUAcceleratorConfig as CPUAcceleratorConfig
from nshtrainer.trainer.accelerator import (
    CUDAAcceleratorConfig as CUDAAcceleratorConfig,
)
from nshtrainer.trainer.accelerator import MPSAcceleratorConfig as MPSAcceleratorConfig
from nshtrainer.trainer.accelerator import XLAAcceleratorConfig as XLAAcceleratorConfig
from nshtrainer.trainer.accelerator import accelerator_registry as accelerator_registry

__all__ = [
    "AcceleratorConfig",
    "AcceleratorConfigBase",
    "CPUAcceleratorConfig",
    "CUDAAcceleratorConfig",
    "MPSAcceleratorConfig",
    "XLAAcceleratorConfig",
    "accelerator_registry",
]



from __future__ import annotations

import logging
from typing import Literal, cast

import torch
import torch.nn as nn
from lightning.pytorch import Callback, LightningModule, Trainer
from torch.optim import Optimizer
from typing_extensions import final, override

from .base import CallbackConfigBase, callback_registry

log = logging.getLogger(__name__)


@final
@callback_registry.register
class NormLoggingCallbackConfig(CallbackConfigBase):
    name: Literal["norm_logging"] = "norm_logging"

    log_grad_norm: bool | str | float = False
    """If enabled, will log the gradient norm (averaged across all model parameters) to the logger."""
    log_grad_norm_per_param: bool | str | float = False
    """If enabled, will log the gradient norm for each model parameter to the logger."""

    log_param_norm: bool | str | float = False
    """If enabled, will log the parameter norm (averaged across all model parameters) to the logger."""
    log_param_norm_per_param: bool | str | float = False
    """If enabled, will log the parameter norm for each model parameter to the logger."""

    def __bool__(self):
        return any(
            v
            for v in (
                self.log_grad_norm,
                self.log_grad_norm_per_param,
                self.log_param_norm,
                self.log_param_norm_per_param,
            )
        )

    @override
    def create_callbacks(self, trainer_config):
        if not self:
            return

        yield NormLoggingCallback(self)


def grad_norm(
    module: nn.Module,
    norm_type: float | int | str,
    group_separator: str = "/",
    grad: bool = True,
) -> dict[str, torch.Tensor | float]:
    """Compute each parameter's gradient's norm and their overall norm.

    The overall norm is computed over all gradients together, as if they
    were concatenated into a single vector.

    Args:
        module: :class:`torch.nn.Module` to inspect.
        norm_type: The type of the used p-norm, cast to float if necessary.
            Can be ``'inf'`` for infinity norm.
        group_separator: The separator string used by the logger to group
            the gradients norms in their own subfolder instead of the logs one.

    Return:
        norms: The dictionary of p-norms of each parameter's gradient and
            a special entry for the total p-norm of the gradients viewed
            as a single vector.
    """
    norm_type = float(norm_type)
    if norm_type <= 0:
        raise ValueError(
            f"`norm_type` must be a positive number or 'inf' (infinity norm). Got {norm_type}"
        )

    if grad:
        norms = {
            f"grad_{norm_type}_norm{group_separator}{name}": p.grad.data.norm(norm_type)
            for name, p in module.named_parameters()
            if p.grad is not None
        }
        if norms:
            total_norm = torch.tensor(list(norms.values())).norm(norm_type)
            norms[f"grad_{norm_type}_norm_total"] = total_norm
    else:
        norms = {
            f"param_{norm_type}_norm{group_separator}{name}": p.data.norm(norm_type)
            for name, p in module.named_parameters()
            if p.grad is not None
        }
        if norms:
            total_norm = torch.tensor(list(norms.values())).norm(norm_type)
            norms[f"param_{norm_type}_norm_total"] = total_norm

    return norms


def _to_norm_type(log_grad_norm_per_param: float | str | Literal[True]):
    norm_type = 2.0
    if log_grad_norm_per_param is not True:
        norm_type = log_grad_norm_per_param
    return norm_type


def compute_norm(
    pl_module: LightningModule,
    optimizer: Optimizer | None = None,
    p: float | str = 2.0,
    *,
    grad: bool,
) -> torch.Tensor:
    if optimizer is not None:
        tensors = [
            cast(torch.Tensor, p.grad if grad else p)
            for group in optimizer.param_groups
            for p in group["params"]
            if p.grad is not None
        ]
    else:
        tensors = [
            p.grad if grad else p for p in pl_module.parameters() if p.grad is not None
        ]

    if not tensors:
        return torch.tensor(0.0, device=pl_module.device)

    return torch.norm(torch.stack([torch.norm(g, p=p) for g in tensors]), p=p)


class NormLoggingCallback(Callback):
    def __init__(self, config: "NormLoggingCallbackConfig"):
        super().__init__()

        self.config = config

    def _perform_norm_logging(
        self,
        pl_module: LightningModule,
        optimizer: Optimizer,
        prefix: str,
    ):
        # Gradient norm logging
        if log_grad_norm := self.config.log_grad_norm:
            norm = compute_norm(
                pl_module,
                optimizer,
                _to_norm_type(log_grad_norm),
                grad=True,
            )
            pl_module.log(f"{prefix}grad_norm", norm, on_step=True, on_epoch=False)
        if log_grad_norm_per_param := self.config.log_grad_norm_per_param:
            norm_type = _to_norm_type(log_grad_norm_per_param)
            pl_module.log_dict(
                {
                    f"{prefix}{k}": v
                    for k, v in grad_norm(pl_module, norm_type, grad=True).items()
                }
            )

        # Parameter norm logging
        if log_param_norm := self.config.log_param_norm:
            norm = compute_norm(
                pl_module,
                optimizer,
                _to_norm_type(log_param_norm),
                grad=False,
            )
            pl_module.log(f"{prefix}param_norm", norm, on_step=True, on_epoch=False)
        if log_param_norm_per_param := self.config.log_param_norm_per_param:
            norm_type = _to_norm_type(log_param_norm_per_param)
            pl_module.log_dict(
                {
                    f"{prefix}{k}": v
                    for k, v in grad_norm(pl_module, norm_type, grad=False).items()
                }
            )

    @override
    def on_after_backward(self, trainer: Trainer, pl_module: LightningModule):
        if len(trainer.optimizers) == 1:
            optimizer = trainer.optimizers[0]
            self._perform_norm_logging(pl_module, optimizer, prefix="train/")
        else:
            for i, optimizer in enumerate(trainer.optimizers):
                self._perform_norm_logging(
                    pl_module, optimizer, prefix=f"train/optimizer_{i}/"
                )



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

import nshconfig as C
from lightning.pytorch.loggers import Logger

if TYPE_CHECKING:
    from ..trainer._config import TrainerConfig


class LoggerConfigBase(C.Config, ABC):
    enabled: bool = True
    """Enable this logger."""

    priority: int = 0
    """Priority of the logger. Higher priority loggers are created first."""

    log_dir: C.DirectoryPath | None = None
    """Directory to save the logs to. If None, will use the default log directory for the trainer."""

    @abstractmethod
    def create_logger(self, trainer_config: TrainerConfig) -> Logger | None: ...

    def disable_(self):
        self.enabled = False
        return self

    def __bool__(self):
        return self.enabled

    def resolve_logger_dirname(self) -> str:
        if not (name := getattr(self, "name", None)):
            raise ValueError(
                "Logger must have a name attribute to resolve the directory name.\n"
                "Otherwise, you must override `resolve_logger_dirname`."
            )

        return name


logger_registry = C.Registry(LoggerConfigBase, discriminator="name")



from __future__ import annotations

from typing import Annotated

import nshconfig as C
from typing_extensions import TypeAliasType

from ._base import BaseProfilerConfig as BaseProfilerConfig
from .advanced import AdvancedProfilerConfig as AdvancedProfilerConfig
from .pytorch import PyTorchProfilerConfig as PyTorchProfilerConfig
from .simple import SimpleProfilerConfig as SimpleProfilerConfig

ProfilerConfig = TypeAliasType(
    "ProfilerConfig",
    Annotated[
        SimpleProfilerConfig | AdvancedProfilerConfig | PyTorchProfilerConfig,
        C.Field(discriminator="name"),
    ],
)



from __future__ import annotations

from . import transform as dataset_transform
from .balanced_batch_sampler import BalancedBatchSampler as BalancedBatchSampler
from .datamodule import LightningDataModuleBase as LightningDataModuleBase

_ = dataset_transform



from __future__ import annotations

__codegen__ = True

from nshtrainer.metrics import MetricConfig as MetricConfig

from . import _config as _config

__all__ = [
    "MetricConfig",
    "_config",
]



from __future__ import annotations

__codegen__ = True

from nshtrainer.loggers.tensorboard import LoggerConfigBase as LoggerConfigBase
from nshtrainer.loggers.tensorboard import (
    TensorboardLoggerConfig as TensorboardLoggerConfig,
)
from nshtrainer.loggers.tensorboard import logger_registry as logger_registry

__all__ = [
    "LoggerConfigBase",
    "TensorboardLoggerConfig",
    "logger_registry",
]



from __future__ import annotations

__codegen__ = True

from nshtrainer.trainer.trainer import AcceleratorConfigBase as AcceleratorConfigBase
from nshtrainer.trainer.trainer import (
    DistributedPredictionWriterConfig as DistributedPredictionWriterConfig,
)
from nshtrainer.trainer.trainer import EnvironmentConfig as EnvironmentConfig
from nshtrainer.trainer.trainer import StrategyConfigBase as StrategyConfigBase
from nshtrainer.trainer.trainer import TrainerConfig as TrainerConfig

__all__ = [
    "AcceleratorConfigBase",
    "DistributedPredictionWriterConfig",
    "EnvironmentConfig",
    "StrategyConfigBase",
    "TrainerConfig",
]



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

import nshconfig as C
from lightning.fabric.plugins import CheckpointIO, ClusterEnvironment
from lightning.pytorch.plugins.layer_sync import LayerSync
from lightning.pytorch.plugins.precision.precision import Precision
from typing_extensions import TypeAliasType

if TYPE_CHECKING:
    from .._config import TrainerConfig


Plugin = TypeAliasType(
    "Plugin", Precision | ClusterEnvironment | CheckpointIO | LayerSync
)


class PluginConfigBase(C.Config, ABC):
    @abstractmethod
    def create_plugin(self, trainer_config: "TrainerConfig") -> Plugin: ...


plugin_registry = C.Registry(PluginConfigBase, discriminator="name")



from __future__ import annotations

import logging
from typing import Any, Literal, Protocol, runtime_checkable

import torch
import torchmetrics
from lightning.pytorch import Callback, LightningModule, Trainer
from torch.optim import Optimizer
from typing_extensions import final, override

from .base import CallbackConfigBase, callback_registry
from .norm_logging import compute_norm

log = logging.getLogger(__name__)


@final
@callback_registry.register
class GradientSkippingCallbackConfig(CallbackConfigBase):
    name: Literal["gradient_skipping"] = "gradient_skipping"

    threshold: float
    """Threshold to use for gradient skipping."""

    norm_type: str | float = 2.0
    """Norm type to use for gradient skipping."""

    start_after_n_steps: int | None = 100
    """Number of steps to wait before starting gradient skipping."""

    skip_non_finite: bool = False
    """
    If False, it doesn't skip steps with non-finite norms. This is useful when using AMP, as AMP checks for NaN/Inf grads to adjust the loss scale. Otherwise, skips steps with non-finite norms.

    Should almost always be False, especially when using AMP (unless you know what you're doing!).
    """

    @override
    def create_callbacks(self, trainer_config):
        yield GradientSkippingCallback(self)


@runtime_checkable
class HasGradSkippedSteps(Protocol):
    grad_skipped_steps: Any


class GradientSkippingCallback(Callback):
    def __init__(self, config: GradientSkippingCallbackConfig):
        super().__init__()
        self.config = config

    @override
    def setup(self, trainer: Trainer, pl_module: LightningModule, stage: str) -> None:
        if not isinstance(pl_module, HasGradSkippedSteps):
            pl_module.grad_skipped_steps = torchmetrics.SumMetric()

    @override
    def on_before_optimizer_step(
        self, trainer: Trainer, pl_module: LightningModule, optimizer: Optimizer
    ):
        # This should never happen, but just in case
        if not isinstance(pl_module, HasGradSkippedSteps):
            raise TypeError(
                f"Expected HasGradSkippedSteps, got {type(pl_module)} instead"
            )

        # Skip the step if the global step is less than the start_after_n_steps
        # This is because we want to let AMP adjust the loss scale before we start
        if (
            self.config.start_after_n_steps is not None
            and pl_module.global_step < self.config.start_after_n_steps
        ):
            return

        norm = compute_norm(pl_module, optimizer, self.config.norm_type, grad=True)

        # If the norm is NaN/Inf, we don't want to skip the step
        # beacuse AMP checks for NaN/Inf grads to adjust the loss scale.
        if self.config.skip_non_finite and not torch.isfinite(norm).all():
            optimizer.zero_grad()
            pl_module.grad_skipped_steps(1)
            log.warning(
                f"Skipping step at global step {pl_module.global_step} with non-finite norm {norm:.2f}"
            )
        elif (norm > self.config.threshold).any():
            optimizer.zero_grad()
            pl_module.grad_skipped_steps(1)
            log.warning(
                f"Skipping step at global step {pl_module.global_step} with norm {norm:.2f} > {self.config.threshold:.2f}"
            )
        else:
            pl_module.grad_skipped_steps(0)

        pl_module.log(
            "train/grad_skipped_steps",
            pl_module.grad_skipped_steps,
            on_step=True,
            on_epoch=False,
        )



from __future__ import annotations

__codegen__ = True

from nshtrainer.trainer._config import AcceleratorConfig as AcceleratorConfig
from nshtrainer.trainer._config import ActSaveLoggerConfig as ActSaveLoggerConfig
from nshtrainer.trainer._config import (
    BestCheckpointCallbackConfig as BestCheckpointCallbackConfig,
)
from nshtrainer.trainer._config import CallbackConfig as CallbackConfig
from nshtrainer.trainer._config import CallbackConfigBase as CallbackConfigBase
from nshtrainer.trainer._config import (
    CheckpointCallbackConfig as CheckpointCallbackConfig,
)
from nshtrainer.trainer._config import CheckpointSavingConfig as CheckpointSavingConfig
from nshtrainer.trainer._config import CSVLoggerConfig as CSVLoggerConfig
from nshtrainer.trainer._config import (
    DebugFlagCallbackConfig as DebugFlagCallbackConfig,
)
from nshtrainer.trainer._config import DirectoryConfig as DirectoryConfig
from nshtrainer.trainer._config import (
    EarlyStoppingCallbackConfig as EarlyStoppingCallbackConfig,
)
from nshtrainer.trainer._config import EnvironmentConfig as EnvironmentConfig
from nshtrainer.trainer._config import GradientClippingConfig as GradientClippingConfig
from nshtrainer.trainer._config import HuggingFaceHubConfig as HuggingFaceHubConfig
from nshtrainer.trainer._config import (
    LastCheckpointCallbackConfig as LastCheckpointCallbackConfig,
)
from nshtrainer.trainer._config import (
    LearningRateMonitorConfig as LearningRateMonitorConfig,
)
from nshtrainer.trainer._config import LogEpochCallbackConfig as LogEpochCallbackConfig
from nshtrainer.trainer._config import LoggerConfig as LoggerConfig
from nshtrainer.trainer._config import LoggerConfigBase as LoggerConfigBase
from nshtrainer.trainer._config import MetricConfig as MetricConfig
from nshtrainer.trainer._config import (
    MetricValidationCallbackConfig as MetricValidationCallbackConfig,
)
from nshtrainer.trainer._config import (
    NormLoggingCallbackConfig as NormLoggingCallbackConfig,
)
from nshtrainer.trainer._config import (
    OnExceptionCheckpointCallbackConfig as OnExceptionCheckpointCallbackConfig,
)
from nshtrainer.trainer._config import PluginConfig as PluginConfig
from nshtrainer.trainer._config import ProfilerConfig as ProfilerConfig
from nshtrainer.trainer._config import (
    RLPSanityChecksCallbackConfig as RLPSanityChecksCallbackConfig,
)
from nshtrainer.trainer._config import SanityCheckingConfig as SanityCheckingConfig
from nshtrainer.trainer._config import (
    SharedParametersCallbackConfig as SharedParametersCallbackConfig,
)
from nshtrainer.trainer._config import StrategyConfig as StrategyConfig
from nshtrainer.trainer._config import (
    TensorboardLoggerConfig as TensorboardLoggerConfig,
)
from nshtrainer.trainer._config import TrainerConfig as TrainerConfig
from nshtrainer.trainer._config import WandbLoggerConfig as WandbLoggerConfig

__all__ = [
    "AcceleratorConfig",
    "ActSaveLoggerConfig",
    "BestCheckpointCallbackConfig",
    "CSVLoggerConfig",
    "CallbackConfig",
    "CallbackConfigBase",
    "CheckpointCallbackConfig",
    "CheckpointSavingConfig",
    "DebugFlagCallbackConfig",
    "DirectoryConfig",
    "EarlyStoppingCallbackConfig",
    "EnvironmentConfig",
    "GradientClippingConfig",
    "HuggingFaceHubConfig",
    "LastCheckpointCallbackConfig",
    "LearningRateMonitorConfig",
    "LogEpochCallbackConfig",
    "LoggerConfig",
    "LoggerConfigBase",
    "MetricConfig",
    "MetricValidationCallbackConfig",
    "NormLoggingCallbackConfig",
    "OnExceptionCheckpointCallbackConfig",
    "PluginConfig",
    "ProfilerConfig",
    "RLPSanityChecksCallbackConfig",
    "SanityCheckingConfig",
    "SharedParametersCallbackConfig",
    "StrategyConfig",
    "TensorboardLoggerConfig",
    "TrainerConfig",
    "WandbLoggerConfig",
]



from __future__ import annotations

__codegen__ = True

from nshtrainer.lr_scheduler.reduce_lr_on_plateau import EpochsConfig as EpochsConfig
from nshtrainer.lr_scheduler.reduce_lr_on_plateau import (
    LRSchedulerConfigBase as LRSchedulerConfigBase,
)
from nshtrainer.lr_scheduler.reduce_lr_on_plateau import MetricConfig as MetricConfig
from nshtrainer.lr_scheduler.reduce_lr_on_plateau import (
    ReduceLROnPlateauConfig as ReduceLROnPlateauConfig,
)
from nshtrainer.lr_scheduler.reduce_lr_on_plateau import (
    lr_scheduler_registry as lr_scheduler_registry,
)

__all__ = [
    "EpochsConfig",
    "LRSchedulerConfigBase",
    "MetricConfig",
    "ReduceLROnPlateauConfig",
    "lr_scheduler_registry",
]



from __future__ import annotations

__codegen__ = True

from nshtrainer.callbacks import ActSaveConfig as ActSaveConfig
from nshtrainer.callbacks import (
    BestCheckpointCallbackConfig as BestCheckpointCallbackConfig,
)
from nshtrainer.callbacks import CallbackConfig as CallbackConfig
from nshtrainer.callbacks import CallbackConfigBase as CallbackConfigBase
from nshtrainer.callbacks import DebugFlagCallbackConfig as DebugFlagCallbackConfig
from nshtrainer.callbacks import (
    DirectorySetupCallbackConfig as DirectorySetupCallbackConfig,
)
from nshtrainer.callbacks import (
    DistributedPredictionWriterConfig as DistributedPredictionWriterConfig,
)
from nshtrainer.callbacks import (
    EarlyStoppingCallbackConfig as EarlyStoppingCallbackConfig,
)
from nshtrainer.callbacks import EMACallbackConfig as EMACallbackConfig
from nshtrainer.callbacks import EpochTimerCallbackConfig as EpochTimerCallbackConfig
from nshtrainer.callbacks import (
    FiniteChecksCallbackConfig as FiniteChecksCallbackConfig,
)
from nshtrainer.callbacks import (
    GradientSkippingCallbackConfig as GradientSkippingCallbackConfig,
)
from nshtrainer.callbacks import (
    LastCheckpointCallbackConfig as LastCheckpointCallbackConfig,
)
from nshtrainer.callbacks import LearningRateMonitorConfig as LearningRateMonitorConfig
from nshtrainer.callbacks import LogEpochCallbackConfig as LogEpochCallbackConfig
from nshtrainer.callbacks import (
    MetricValidationCallbackConfig as MetricValidationCallbackConfig,
)
from nshtrainer.callbacks import NormLoggingCallbackConfig as NormLoggingCallbackConfig
from nshtrainer.callbacks import (
    OnExceptionCheckpointCallbackConfig as OnExceptionCheckpointCallbackConfig,
)
from nshtrainer.callbacks import (
    PrintTableMetricsCallbackConfig as PrintTableMetricsCallbackConfig,
)
from nshtrainer.callbacks import (
    RLPSanityChecksCallbackConfig as RLPSanityChecksCallbackConfig,
)
from nshtrainer.callbacks import (
    SharedParametersCallbackConfig as SharedParametersCallbackConfig,
)
from nshtrainer.callbacks import (
    WandbUploadCodeCallbackConfig as WandbUploadCodeCallbackConfig,
)
from nshtrainer.callbacks import WandbWatchCallbackConfig as WandbWatchCallbackConfig
from nshtrainer.callbacks import callback_registry as callback_registry
from nshtrainer.callbacks.checkpoint._base import (
    BaseCheckpointCallbackConfig as BaseCheckpointCallbackConfig,
)
from nshtrainer.callbacks.checkpoint._base import (
    CheckpointMetadata as CheckpointMetadata,
)
from nshtrainer.callbacks.early_stopping import MetricConfig as MetricConfig

from . import actsave as actsave
from . import base as base
from . import checkpoint as checkpoint
from . import debug_flag as debug_flag
from . import directory_setup as directory_setup
from . import distributed_prediction_writer as distributed_prediction_writer
from . import early_stopping as early_stopping
from . import ema as ema
from . import finite_checks as finite_checks
from . import gradient_skipping as gradient_skipping
from . import log_epoch as log_epoch
from . import lr_monitor as lr_monitor
from . import metric_validation as metric_validation
from . import norm_logging as norm_logging
from . import print_table as print_table
from . import rlp_sanity_checks as rlp_sanity_checks
from . import shared_parameters as shared_parameters
from . import timer as timer
from . import wandb_upload_code as wandb_upload_code
from . import wandb_watch as wandb_watch

__all__ = [
    "ActSaveConfig",
    "BaseCheckpointCallbackConfig",
    "BestCheckpointCallbackConfig",
    "CallbackConfig",
    "CallbackConfigBase",
    "CheckpointMetadata",
    "DebugFlagCallbackConfig",
    "DirectorySetupCallbackConfig",
    "DistributedPredictionWriterConfig",
    "EMACallbackConfig",
    "EarlyStoppingCallbackConfig",
    "EpochTimerCallbackConfig",
    "FiniteChecksCallbackConfig",
    "GradientSkippingCallbackConfig",
    "LastCheckpointCallbackConfig",
    "LearningRateMonitorConfig",
    "LogEpochCallbackConfig",
    "MetricConfig",
    "MetricValidationCallbackConfig",
    "NormLoggingCallbackConfig",
    "OnExceptionCheckpointCallbackConfig",
    "PrintTableMetricsCallbackConfig",
    "RLPSanityChecksCallbackConfig",
    "SharedParametersCallbackConfig",
    "WandbUploadCodeCallbackConfig",
    "WandbWatchCallbackConfig",
    "actsave",
    "base",
    "callback_registry",
    "checkpoint",
    "debug_flag",
    "directory_setup",
    "distributed_prediction_writer",
    "early_stopping",
    "ema",
    "finite_checks",
    "gradient_skipping",
    "log_epoch",
    "lr_monitor",
    "metric_validation",
    "norm_logging",
    "print_table",
    "rlp_sanity_checks",
    "shared_parameters",
    "timer",
    "wandb_upload_code",
    "wandb_watch",
]
